/*
  Questo programma calcola le regioni del grafo del Web
  (si assume che il grafo sia conforme al modello del WebGraph).
  In particolare viene calcolata la massima componente fortemente connessa (regione SCC),
  e le regioni IN, OUT, TENDRILS, DISCONNECTED.
  Della regione TENDRILS viene inoltre determinato il sottoinsieme TUBES.
  Il programma sfrutta la visita di raggiungibilita' con due bit d'informazione per nodo.

  La SCC massima viene calcolata per tentativi: si estraggono una serie di nodi a caso,
  per ciascuno di essi se ne calcola la corrispondente SCC mediante una doppia visita (in avanti
  e all'indietro), e alla fine si restituisce la componente piu' grande tra quelle trovate.
  Con un numero adeguato di tentativi, e se il grafo e' appunto del modello del WebGraph, la
  probabilita' di aver calcolato proprio la SCC massima diventa molto elevata.

  L'algoritmo e' semi-external come la visita di raggiungibilita' impiegata (sono richiesti 2 bit
  per nodo). I vettori di bits temporanei (ad es. la piu' grande SCC finora calcolata, o l'unione
  delle SCC gia' incontrate) sono salvati su disco.

  Vengono fornite alcune opzioni per limitare il numero di tentativi, arrestando il processo quando
  la percentuale di nodi di tutte le SCC calcolate supera una certa soglia, o quando sono state
  calcolate abbastanza SCC, oppure quando e' trascorso gia' un certo tempo.

  Le regioni IN, OUT, TENDRILS, DISCONNECTED si ottengono con delle semplici visite, una volta
  che SCC e' nota. In particolare:
  SCC --> IN, OUT
  IN, OUT [,SCC] --> TENDRILS (con TUBES)
  SCC,IN,OUT,TENDRILS --> DISCONNECTED


  Di seguito viene riportata la descrizione dell'algoritmo di visita che e' stato impiegato.


  -----Algoritmo di visita (calcolo della raggiungibilita')----------------------------------

  La visita calcola l'insieme dei nodi raggiungibili in un grafo a partire da uno
  o piu' nodi di partenza. L'ordine di visita non e' fissato (non e' ad es. ne' una BFS,
  ne' una DFS). Viene allocato un blocco di memoria in cui viene caricata ogni volta una porzione 
  del grafo di input e su di essa viene effettuata una visita parziale.
  Una visita dell'intero grafo comporta una o piu' scansioni dei file di input, e ciascuna
  di esse determina l'elaborazione in sequenza dei vari blocchi.

  La porzione di grafo viene rappresentata con un blocco di memoria contenente le liste di
  adiacenza, con un opportuno dimensionamento del vettore dei puntatori e del vettore dei
  nodi adiacenti. Le liste di adiacenza contengono ogni volta i dati relativi ad un intervallo
  contiguo di nodi del grafo. L'intervallo e' in generale di tipo circolare, cioe', se si arriva
  alla fine dei file del grafo e vi e' ancora spazio libero in memoria, si continua con il
  caricamento dei nodi iniziali.

  La visita e' semi-external perche' devono essere mantenute contemporaneamente in memoria
  due informazioni su ciascun nodo:
  -un bit che indica se il nodo e' stato finora raggiunto dalla visita
  -un bit che indica se e' stato completato il processamento dell'intera lista di adiacenza del nodo nel grafo
  Ci saranno quindi due vettori di bit in memoria, realizzati con la classe VettoreBits.

  Per ogni blocco del grafo caricato in memoria:
  1) si cercano i nodi raggiunti e non-completati
  2) se ne esamina la lista di adiacenza
  3) ogni adiacente diventa raggiunto
  4) il nodo, una volta ultimato il processamento della sua lista di adiacenza, diventa completato.

  La ricerca dei nodi raggiunti e non-completati avviene prima sequenzialmente (con l'apposita
  funzione ottimizzata di VettoreBits). Si ripassa sui nodi raggiunti e non-completati, che sono
  stati prodotti esaminando le liste di adiacenza del blocco attuale, avendoli messi in una pila
  (la visita di un blocco e' vagamente simile ad una DFS).

  Alla fine dell'elaborazione di ciascun blocco non rimangono nodi del blocco raggiunti e
  non-completati. Alla fine dell'intera visita nessun nodo del grafo e' raggiunto e
  non-completato. Tutti i nodi marcati come raggiunti costituiscono il risultato della visita.

  Sono previste ottimizzazioni per evitare di caricare nelle liste di adiacenza informazioni
  che non verrebbero utilizzate, ovvero quelle relative a nodi gia' completati.


  Ernesto Ricci
   7/4/2003
*/



#include <getopt.h>
#include "main.h"
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include "comuni.h"
#include "../common/scc_utl/VettoreBits.h"





//nomi dei file temporanei usati:
#ifdef USA_DIR_TMP
  const char nome_filetmp_max_SCC[]= "/tmp/file_SCCmax_tmp";
  const char nome_filetmp_tutte_le_SCC[]= "/tmp/file_allSCCs_tmp";
  const char nome_file_tmp_IN[]= "/tmp/file_IN_tmp";
  const char nome_file_tmp_OUT[]= "/tmp/file_OUT_tmp";
  const char nome_file_tmp_TENDRILS_lato_IN[]= "/tmp/file_TENDRILS_IN_tmp";
#else
  const char nome_filetmp_max_SCC[]= "file_SCCmax_tmp";
  const char nome_filetmp_tutte_le_SCC[]= "file_allSCCs_tmp";
  const char nome_file_tmp_IN[]= "file_IN_tmp";
  const char nome_file_tmp_OUT[]= "file_OUT_tmp";
  const char nome_file_tmp_TENDRILS_lato_IN[]= "file_TENDRILS_IN_tmp";
#endif


unsigned long numero_nodi; //numero nodi del grafo
unsigned long numero_archi; //numero archi del grafo

unsigned long size_Set_IN= 0;
unsigned long size_Set_OUT= 0;

bool visita_avanti; //indica se correntemente operiamo su i successori (visita in avanti) o sui predecessori (visita all'indietro)
char * grafoADIAC; //seleziona l'informazione corrispondente in base al tipo di visita
long numADIAC; //seleziona l'informazione corrispondente in base al tipo di visita


//opzioni riga di comando:
char grafoINFO[1000];
char grafoPRED[1000];
char grafoSUCC[1000];
char * nome_mfile_output;
bool genera_nodi_random;
char * nome_mfile_nodi_test;
unsigned long num_lanci_da_effettuare; //indica quanti nodi visitare per trovare la SCC grande
long long memoria_max= -1; //memoria massima cha il programma puo' usare
bool esegui_comunque; //flag per eseguire incondizionatamente la visita anche se la memoria necessaria ?maggiore di quella disponibile
bool calcola_IN;
bool calcola_OUT;
bool calcola_TENDRILS;
bool calcola_DISC;
bool salva_tutte_le_SCC;
char * nome_mfile_set_IN;
char * nome_mfile_set_OUT;
char * nome_mfile_set_TENDRILS;
char * nome_mfile_set_DISC;
bool salva_output_come_bits;
bool usato_seed;
bool specificato_max_tempo;
bool specificato_max_perc;
bool specificato_max_num_SCC;
bool tentativi_random_illimitati;
unsigned long seed;
unsigned long tempo_max;
unsigned long numSCC_max;
unsigned long percNodi_max;


bool richiesta_controlc; //mem. una richiesta di uscire con CTRL+C
time_t tempo_inizio;

//numero parti dei vari multifile di input:
long numINFO; //numero parti multifile dei nodi
long numPRED; //numero parti multifile predecessori
long numSUCC; //numero parti multifile successori
long numFileTentativi; //numero parti multifile con l'elenco dei nodi da provare


//puntatori agli oggetti per la lettura degli archi:
multiFileReader * mf_info= NULL;
multiFileReader * mf_adiac= NULL;
BufferLetturaMultifile<struct grafoInfo> * buf_info= NULL;
BufferLetturaMultifile<unsigned long> * buf_adiac= NULL;



//puntatori alle altre strutture in memoria:
unsigned long * blocco_memoria= NULL;

unsigned long * vettore_adiac; //e' un "alias" di blocco_memoria, perche' gli adiacenti vengono mem. nella parte iniziale del vettore

VettoreBits * nodi_completati= NULL;
VettoreBits * nodi_raggiunti= NULL;

//informazioni sulla porzione di grafo mantenuta nel blocco di liste di adiacenza:
unsigned long capienza_blocco_memoria; //numero di elementi del vettore 'blocco_memoria' allocato
unsigned long quanti_nodi_in_memoria; //quanti nodi sono contenuti nella porzioni di grafo mantenuta attualmente in memoria
unsigned long primo_nodo_in_memoria; //il primo nodo dell'intervallo di nodi mantenuto attualmente in memoria
unsigned long ultimo_nodo_in_memoria; //l'ultimo nodo dell'intervallo di nodi mantenuto attualmente in memoria (N.B. L'intervallo di nodi e' "circolare", percio' l'identificatore dell'ultimo nodo dell'intervallo puo' precedere l'identificatore del primo nodo)

unsigned long ultimo_nodo_considerato; //ultimo nodo di cui e' stata letta la info, ma che non e' stato caricato nel blocco (e' il primo da caricare nel blocco successivo)
unsigned long archi_ultimo_nodo_considerato; //degree del nodo
long long punt_adiac_ultimo_nodo_considerato; //puntatore agli adiacenti del nodo

unsigned long nodi_ancora_da_completare; //quanti nodi sono stati raggiunti, ma non sono ancora stati espansi




//scrive un messaggio ed esce dopo liberando le strutture di memoria impegnate
//per quanto riguarda i parametri usa la stessa sintassi di printf:  printf(const char *formato, ...)
void gestione_errore(const char *formato, ...)
{
  va_list ap;
  //stampiamo il messaggio d'errore ed usciamo:
  fprintf(stderr,"\n");
  va_start (ap, formato); //inizializziamo la struttura 'ap'
  vfprintf (stderr, formato, ap); //scriviamo il messaggio d'errore con tutti i parametri eventuali
  va_end (ap); //chiudiamo con la variabile 'ap'
  fprintf(stderr,"\n");
  perror("");

  libera_memoria();
  ABORT();
}


/*
void indiceProgressione(const char* testo)
{
  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  printf("\r [%c] %s\r", cont, testo);
  fflush(stdout);
}


void indiceProgressione(const char* testo,long perc1,long perc2){

  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  if (perc1<0) perc1=0;
  else if (perc1>100) perc1=100;

  if (perc2<0) perc2=0;
  else if (perc2>100) perc2=100;

  perc2=(long) (20.0*perc2/100);
  //indicatore
  char st[50];

  strcpy(st,"____________________");
  for (long i=0;(i<20)&&(i<perc2);i++ )
     st[i]='=';

  //limite a 30 la lunghezza del testo
  char s[55];
  long i=0;
  for (i=0; (testo[i]!=0)&&(i<40);i++)
    s[i]=testo[i];
  for (; (i<40);i++)
    s[i]=' ';
  s[40]=0;


	printf("\r [%c] %3ld%% %20s %s\r",cont,perc1,st, s);
  fflush(stdout);
}
*/


void eraseMultiFile(const char* nomeMulti){

    long conta=0;

    FILE* fp;
    char nomeFile[1000];

    while(true){
      sprintf(nomeFile,nomeMulti,conta);
      fp=fopen(nomeFile,"r");
      if (fp!=NULL){
          conta++;
          fclose(fp);
          remove(nomeFile);
      }
      else
          return;
    }
}



/*
 allochiamo una quantita' di memoria dipendente dalla memoria disponibile:
 1) vediamo se l'intero grafo puo' entrare in memoria
 2) allochiamo tutta la memoria a disposizione per rappresentare il piu' grande blocco possibile di liste di adiacenza
 3) se la memoria dovesse essere scarsa allochiamo un blocco per liste di adiacenza di dimensione minima
*/
void alloca_memoria()
{
  //dimensionamento della memoria da allocare:
  long long memoria_prevista_tot= (long long) 2 * numero_nodi/8 + DIM_BLOCCO_STACK + DIM_BUF_INFO + DIM_BUF_ADIAC + DIM_BUF_LETTURA_FILE_INPUT_NODI + DIM_BUF_SCRITTURA_INFOTREE + DIM_BUF_SCRITTURA_OUTPUT ; //approssimativamente la memoria richiesta

  long long mem_lista_adiacenza_completa= ((long long) 1 + numero_nodi + numero_archi)*sizeof(unsigned long); //memoria per contenere nel blocco tutti i nodi con tutti gli archi
  if(memoria_prevista_tot + mem_lista_adiacenza_completa <= memoria_max) //caso in cui abbiamo memoria in abbondanza
  {
    capienza_blocco_memoria= mem_lista_adiacenza_completa / sizeof(unsigned long);
    memoria_prevista_tot += mem_lista_adiacenza_completa;
  }
  else
  if(memoria_prevista_tot + DIM_MINIMA_BLOCCO_MEMORIA >= memoria_max) //caso in cui sono gia' al di sopra della memoria max
  {
    capienza_blocco_memoria= DIM_MINIMA_BLOCCO_MEMORIA / sizeof(unsigned long);
    memoria_prevista_tot += DIM_MINIMA_BLOCCO_MEMORIA;
  }
  else //caso normale: sfruttiamo tutto il resto della mem. disponibile per il nostro blocco
  {
    long long restante= (memoria_max - memoria_prevista_tot);
    capienza_blocco_memoria= restante / sizeof(unsigned long);
    memoria_prevista_tot += restante;
  }

  if(memoria_prevista_tot > memoria_max)
  {
    printf("Warning: not enough memory(required at least %lld bytes).\n", memoria_prevista_tot);
    if( !esegui_comunque )
    {
      char car;
      printf("Continue?(y/n)\n");

      car= getchar();
      if((car != 'y') && (car != 's'))
      {
        printf("Aborted by user.\n");
        ABORT();
      }
    }
  }
  
#ifdef TEST_POCA_MEM
capienza_blocco_memoria= numero_nodi;

  VERBOSE_PRINTF("\n****\"TEST con poca memoria\" attivato: utilizzeremo %lu locazioni per il blocco*****\n", capienza_blocco_memoria);
#endif

  //allocazione vera e propria delle strutture di memoria:
  nodi_completati= new VettoreBits(numero_nodi);
  nodi_raggiunti= new VettoreBits(numero_nodi);
  blocco_memoria= new unsigned long[capienza_blocco_memoria];
  accertati_che( (nodi_completati != NULL) && (nodi_raggiunti != NULL) && (blocco_memoria != NULL),"Errore nell'allocazione della memoria.");
  vettore_adiac= blocco_memoria; //e' la parte iniziale del vettore
  
#ifdef VERBOSE_DEBUG
  printf("\n> Memoria allocata: circa %lld KB\n> Dimensione del blocco per la lista di adiacenza: %lld KB\n\n", memoria_prevista_tot/1024, ((long long) capienza_blocco_memoria) * sizeof(unsigned long) / 1024 + 1);
#endif
}



//chiude e dealloca i buffer per la lettura del grafo
void dealloca_buffers_lettura()
{
  if(mf_info != NULL)
  {
    mf_info->close();
    delete mf_info;
    mf_info= NULL;
  }

  if(mf_adiac != NULL)
  {
    mf_adiac->close();
    delete mf_adiac;
    mf_adiac= NULL;
  }

  if(buf_info != NULL)
  {
    delete buf_info;
    buf_info= NULL;
  }

  if(buf_adiac != NULL)
  {
    delete buf_adiac;
    buf_adiac= NULL;
  }
}



//dealloca ogni oggetto che occupa memoria
void libera_memoria()
{
  if(blocco_memoria != NULL)
  {
    delete blocco_memoria;
    blocco_memoria= NULL;
  }

  if(nodi_raggiunti != NULL)
  {
    delete nodi_raggiunti;
    nodi_raggiunti= NULL;
  }

  if(nodi_completati != NULL)
  {
    delete nodi_completati;
    nodi_completati= NULL;
  }

  dealloca_buffers_lettura();
}



//N.B. dopo che viene cambiata la direzione va effettuato il rewind sui multifile
void imposta_direzione_visita(bool in_avanti)
{
  visita_avanti= in_avanti;

  if(in_avanti)
  {
    grafoADIAC= grafoSUCC;
    numADIAC= numSUCC;
  }
  else
  {
    grafoADIAC= grafoPRED;
    numADIAC= numPRED;
  }
}



//chiude, dealloca, rialloca e riapre i multifile per poter ricominciare a leggere dall'inizio.
void rewind_multifile()
{
  dealloca_buffers_lettura();
  
  mf_info= new multiFileReader(grafoINFO, numINFO); //apertura dei multifile di input
  mf_adiac= new multiFileReader(grafoADIAC, numADIAC);
  accertati_che( (mf_info!=NULL)&&(mf_adiac!=NULL),"memory allocation error (1)");
  //buffer per leggere il multifile a blocchi:
  buf_info= new BufferLetturaMultifile<struct grafoInfo>(mf_info, DIM_BUF_INFO); //buffer per leggere il multifile a blocchi
  buf_adiac= new BufferLetturaMultifile<unsigned long>(mf_adiac, DIM_BUF_ADIAC);
  accertati_che( (buf_info!=NULL)&&(buf_adiac!=NULL),"memory allocation error for buffers");
}



/*
  si posiziona con i buffers di lettura sui dati realtivi al nodo specificato in 'nodo_destinazione'
  se riavvolgi_comunque= true viene effettuato come prima cosa il rewiring dei buffers, altrimenti questo si fara' solo se necessario, oppure si continueranno ad usare i buffer gia' allocati
*/
void salta_al_nodo(unsigned long nodo_destinazione, bool riavvolgi_comunque= false)
{
  bool ok;
  long long pos_adiac_dove_saltare;
  struct grafoInfo info;  
  bool usato_rewind= false;
  
  if((nodo_destinazione < ultimo_nodo_considerato) || riavvolgi_comunque) //e' necessario fare il rewind?
  {  
    rewind_multifile(); //apriamo i multifile in lettura

    punt_adiac_ultimo_nodo_considerato= 0; //in modo che il salto della skip degl adiacenti sara' assoluto e non relativo
    usato_rewind= true;
  }
  else //non dobbiamo necessariamente riavvolgere
    if(nodo_destinazione==ultimo_nodo_considerato) //sono gia' sul punto prescelto
      return; //non devo fare niente

  if(usato_rewind)
    ok= buf_info->skip(nodo_destinazione); //saltiamo le info dei nodi precedenti quello che ci interessa
  else //e ' (nodo_destinazione > ultimo_nodo_considerato): usato_rewind=false --> (nodo_destinazione >= ultimo_nodo_considerato), ma se sono uguali saremmo gia' usciti prima
    ok= buf_info->skip(nodo_destinazione - ultimo_nodo_considerato - 1); //saltiamo le info dei nodi precedenti quello che ci interessa, considerando che la info del nodo 'ultimo_nodo_considerato' e' stata gia' letta
    
  accertati_che( ok, MSG_ERRORE_SEEK, grafoINFO);

  long ris= buf_info->get(info); //leggiamo l'info del nodo che ci interessa
  accertati_che( ris, MSG_ERRORE_LETTURA_MULTIFILE, grafoINFO);

  //simuliamo quello che farebbe leggi_info():
  if(visita_avanti)
  {
    archi_ultimo_nodo_considerato= info.outDeg;
    pos_adiac_dove_saltare= info.puntOut;
  }
  else
  {
    archi_ultimo_nodo_considerato= info.inDeg;
    pos_adiac_dove_saltare= info.puntIn;
  }

  ok= buf_adiac->skip(pos_adiac_dove_saltare - punt_adiac_ultimo_nodo_considerato); //saltiamo gli adiacenti prima di quelli di nodo_destinazione
  accertati_che( ok, MSG_ERRORE_SEEK, grafoADIAC);

  punt_adiac_ultimo_nodo_considerato= pos_adiac_dove_saltare;
  ultimo_nodo_considerato= nodo_destinazione;
}



//funzioni che fanno da "wrapper" alle analoghe di VettoreBits, effettuando un controllo sugli errori
inline bool is_completato(unsigned long nodo)
{
  long errore;
  bool risposta= nodi_completati->get(nodo, errore);
  accertati_che( (errore==0), MSG_ERRORE_BITS);
  return risposta;
}

inline bool is_raggiunto(unsigned long nodo)
{
  long errore;
  bool risposta= nodi_raggiunti->get(nodo, errore);
  accertati_che( (errore==0), MSG_ERRORE_BITS);
  return risposta;
}

inline bool set_completato(unsigned long nodo)
{
  long errore;
  bool risposta= nodi_completati->set(nodo, true, errore);
  accertati_che( (errore==0), MSG_ERRORE_BITS);
  return risposta;
}

inline bool set_raggiunto(unsigned long nodo)
{
  long errore;
  bool risposta= nodi_raggiunti->set(nodo, true, errore);
  accertati_che( (errore==0), MSG_ERRORE_BITS);
  return risposta;
}



//legge circolarmente la prossima info sui nodi settando il numero di adiacenti corrispondenti. Incrementa (in modulo) il conteggio dell'ultimo nodo considerato
void leggi_info()
{
  bool rewind= false;
  if(ultimo_nodo_considerato == numero_nodi - 1) //fine dei nodi
  {
    rewind_multifile();
    rewind= true;
  }
  
  struct grafoInfo info;

  long ris= buf_info->get(info); //leggiamo l'info del nodo successivo
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, grafoINFO);

  if(visita_avanti)
  {
    archi_ultimo_nodo_considerato= info.outDeg;
    punt_adiac_ultimo_nodo_considerato= info.puntOut;
  }
  else
  {
    archi_ultimo_nodo_considerato= info.inDeg;
    punt_adiac_ultimo_nodo_considerato= info.puntIn;
  }

  if(rewind)
    ultimo_nodo_considerato= 0;
  else
    ultimo_nodo_considerato++;
}



//funzione che simula l'accesso ad un vettore normale, in realta' accediamo alla parte finale del blocco_memoria, a partire dalle ultime posizioni verso le prime
inline unsigned long & vett_puntatori(unsigned long X)
{
  return blocco_memoria[capienza_blocco_memoria - 1 - X]; 
}



/*
  cerca, a partire dal nodo 'ultimo_nodo_considerato' compreso, il prossimo nodo raggiunto
  e non-completato del grafo. Se si raggiunge la fine del grafo, la ricerca prosegue
  "riavvolgendosi" dall'inizio. Naturalmente non viene effettuato piu' di un passaggio
  su ogni nodo.
  Ritorna true se e' stato trovato almeno un nodo rispondente, false se e' stato completato
  un intero "giro" tra i nodi senza trovarne nemmeno uno.
*/
bool prossimo_nodo_raggiunto_non_completato(unsigned long & prossimo)
{
  unsigned long last_trovato;
  bool found= nodi_raggiunti->findNext1(last_trovato, ultimo_nodo_considerato);

  while(found) //ricerca tra i nodi seguenti l'ultimo considerato
  {
    if( !is_completato(last_trovato) )
    {
      prossimo= last_trovato;
      return true;
    }
        
    found= nodi_raggiunti->findNext1(last_trovato);
  }
    
  found= nodi_raggiunti->findNext1(last_trovato, 0);
  
  while(found) //ricerca tra i nodi precedenti l'ultimo considerato
  {
    if( !is_completato(last_trovato) )
    {
      prossimo= last_trovato;
      return true;
    }
    else
      if(last_trovato >= ultimo_nodo_considerato) //se abbiamo fatto un "giro"
        return false;

    found= nodi_raggiunti->findNext1(last_trovato);
  }
    
  return false;
}



/*
  cerca, a partire dal nodo 'ultimo_nodo_considerato' compreso, il prossimo nodo non-completato
  del grafo. Se si raggiunge la fine del grafo, la ricerca prosegue
  "riavvolgendosi" dall'inizio. Naturalmente non viene effettuato piu' di un passaggio
  su ogni nodo.
  Ritorna true se e' stato trovato almeno un nodo rispondente, false se e' stato completato
  un intero "giro" tra i nodi senza trovarne nemmeno uno.

  In 'nodi_saltati' viene riportato il numero di nodi da saltare prima di trovare il primo
  nodo non-completato.
*/
inline bool cerca_prossimo_non_completato(unsigned long & prossimo_da_completare, unsigned long & nodi_saltati)
{
  bool found;
  
  found= nodi_completati->findNext0(prossimo_da_completare, ultimo_nodo_considerato); //cerchiamo a partire dal nodo corrente
  if(found) 
    nodi_saltati= prossimo_da_completare - ultimo_nodo_considerato;
  else //non trovato nessun bit = 0 da 'ultimo_nodo_considerato' alla fine
  {
    nodi_saltati= numero_nodi - ultimo_nodo_considerato; //contiamo quelli gia' saltati alla fine
    found= nodi_completati->findNext0(prossimo_da_completare, 0); //riprendiamo a cercare dall'inizio

    if(found)
      nodi_saltati += prossimo_da_completare; //contano come considerati solo quelli precedenti che erano gia' completati e quindi e' come se li avessimo caricati e poi scartati
    else
      nodi_saltati= numero_nodi; //abbiamo fatto almeno un "giro" completo
  }

  return found;
}



/*
  Carica le liste di adiacenza con le informazioni sui prossimi nodi
  da esaminare (a partire da 'ultimo_nodo_considerato' compreso).
  Un unico blocco di memoria ospita i due vettori (puntatori e adiacenti) scrivendo le locazioni
  di uno dall'inizio verso la fine, mentre quelle dell'altro sono scritte dalla fine verso l'inizio.
  Il blocco si riempie quando i due vettori si ricongiungono al centro.

  Non vengono caricati gli adiacenti di nodi gia' completati, ne' gli adiacenti che si riferiscono
  a nodi gia' completati (queste informazioni non verrebbero mai utilizzate).
  Tutti gli eventuali puntatori a nodi completati che dovessero comparire per primi
  nel vettore dei puntatori verrebbero invece saltati.

  Ritorna il numero di nodi che sono stati considerati complessivamente.
  
  PREC: esiste almeno un nodo non completato
*/
unsigned long carica_prossimi_archi()
{
  unsigned long max_nodi_da_considerare= numero_nodi; //per evitare di continuare a ciclare
#ifdef VERBOSE_DEBUG
  printf("Loading edges in memory..."); fflush(stdout);
#endif

  accertati_che((max_nodi_da_considerare<=numero_nodi) && (max_nodi_da_considerare>=1), "Error 1: deve essere\n 1 <= max_nodi_da_considerare <= numero_nodi");
  
  unsigned long nodi_considerati= 0;

  //evitiamo di caricare nodi inutili all'inizio: tali nodi non verrebbero mai utilizzati perche' sono gia' stati completati

  unsigned long primo_da_completare;
  bool found= cerca_prossimo_non_completato(primo_da_completare, nodi_considerati); //i nodi saltati perche' gia' completati vanno contati come se li avessimo effettivamante caricati
  accertati_che( found, "Error 2: Precondizioni di carica_prossimi_archi() non soddisfatte: tutti i nodi erano gia' completati");

  salta_al_nodo(primo_da_completare); //salta nei multifile fino a posizionarsi su 'primo_da_completare' e setta 'ultimo_nodo_considerato'='primo_da_completare'
    
  //a questo punto 'ultimo_nodo_considerato' e' ancora da esaminare

  unsigned long locazioni_free= capienza_blocco_memoria - 1; //una posizione e' riservata per l'ultimo puntatore "nodo fittizio"
  unsigned long num_archi_inseriti= 0;
  unsigned long num_nodi_inseriti= 0;

  accertati_che( (locazioni_free >= 1+archi_ultimo_nodo_considerato), "Error 3: Il nodo %lu ha troppi adiacenti (%lu) per poterli caricare in memoria", ultimo_nodo_considerato, archi_ultimo_nodo_considerato);

  primo_nodo_in_memoria= ultimo_nodo_considerato; //l'ultimo nodo considerato verra' sicuramente inserito (per primo) nel blocco di memoria
  bool continua= true;
  
  //mi serve per fare un ciclo almeno una locazione libera
  do
  {    
    if( is_completato(ultimo_nodo_considerato) ) //se e' completato allora ne ignoriamo la lista di adiacenza
    {
      unsigned long prossimo_da_completare, nodi_saltati;
      cerca_prossimo_non_completato(prossimo_da_completare, nodi_saltati); //ritorna il prossimo non completato a partire da 'ultimo_nodo_considerato' e il numero di nodi saltati a partire da 'ultimo_nodo_considerato' N.B. sappiamo ce ne e' almeno uno ; potremmo ricapitare su uno gia' visto, ma con il conteggio dei nodi considerati non dovremmo avere problemi
      if(nodi_saltati + nodi_considerati >= max_nodi_da_considerare) //con questi abbiamo gia' scandito il numero previsto di nodi
      {
        continua= false;
      }
      else
      if(locazioni_free <= nodi_saltati) //non vi e' piu' spazio nel blocco di memoria: non potremmo inserire nemmeno i puntatori dei nodi saltati
      {
        continua= false;
      }
      else //vi e' ancora spazio sufficiente a inserire i puntatori dei nodi saltati, in modo che potra' essere inserita (se vi e' lo spazio) anche la lista di adiacenza del primo nodo non completato che e' stato trovato
      {
        for(unsigned long i=0; i<nodi_saltati; i++) //riserviamo spazio per i puntatori dei nodi saltati: questi non sono di per se' significativi, ma permettono di mantenere contiguo l'intervallo dei nodi mantenuti in memoria
        {
          vett_puntatori(num_nodi_inseriti)= num_archi_inseriti; //settiamo il puntatore: serve solo una locazione libera
          num_nodi_inseriti++;
        }

        locazioni_free -= nodi_saltati; //consumata una locazione di 'blocco_memoria' per ogni nodo gia' completato (nella parte del vettore dei puntatori)
        salta_al_nodo(prossimo_da_completare); //ci posizioniamo sul prossimo nodo non completato che avevamo trovato
      }

      nodi_considerati += nodi_saltati;
    }
    else //nodo ultimo considerato non ancora completato: ne carichiamo gli adiacenti in memoria se possibile
    {    
      if(locazioni_free >= 1 + archi_ultimo_nodo_considerato) //c'e' sicuramente spazio sufficiente
      {
        vett_puntatori(num_nodi_inseriti)= num_archi_inseriti; //settiamo il puntatore: serve solo una locazione libera
        num_nodi_inseriti++;
        locazioni_free--; //consumata una locazione di 'blocco_memoria' solo per il puntatore
      
        for(unsigned long i=0; i<archi_ultimo_nodo_considerato; i++) //lettura da file della lista di adiacenza del nodo
        {
          long ris= buf_adiac->get( blocco_memoria[num_archi_inseriti] );
          accertati_che( ris, MSG_ERRORE_LETTURA_MULTIFILE, grafoADIAC);
          
          if( !is_completato( blocco_memoria[num_archi_inseriti] ) ) //se invece gia' era completato il nodo destinazione allora e' inutile salvare questo arco: anche se lo abbiamo scritto nella lista di adiacenza non risultera' ed e' come se l'avessimo saltato (l'informazione letta verra' sovrascritta o comunque ignorata)
          {
            num_archi_inseriti++;
            locazioni_free--;
          }
        }

        nodi_considerati++;
        leggi_info(); //passiamo al nodo successivo        
      }
      else //non c'e' abbastanza spazio per inserire l'ultimo nodo considerato con i suoi adiacenti
        continua= false;
    }

  } while(continua && (locazioni_free > 0) && (nodi_considerati < max_nodi_da_considerare));
  
  vett_puntatori(num_nodi_inseriti)= num_archi_inseriti; //"chiudiamo" la lista di adiac.

  quanti_nodi_in_memoria= num_nodi_inseriti;

  if(ultimo_nodo_considerato == 0) //ultimo_nodo_in_memoria deve essere il nodo immediatamente prima (in modulo) di ultimo_nodo_considerato, perche' ultimo_nodo_considerato e' il primo dei nodi dopo l'intervallo dei nodi caricati in memoria
    ultimo_nodo_in_memoria= numero_nodi - 1; //una posizione prima di 0 modulo 'numero_nodi'
  else
    ultimo_nodo_in_memoria= ultimo_nodo_considerato - 1; 
  
#ifdef VERBOSE_DEBUG
  printf(MSG_DONE);
#endif
  
  return nodi_considerati;
}



/*
Ritorna se il nodo specificato e' rappresentato (con la sua lista di adiacenza) nel blocco attualmente in memoria
In memoria ci sono 'quanti_nodi_in_memoria' nodi, dal nodo 'primo_nodo_in_memoria' al nodo 'ultimo_nodo_in_memoria' inclusi.
L'intervallo e' continuo, ma 'ultimo_nodo_in_memoria' puo' essere < 'primo_nodo_in_memoria' se il blocco di nodi si "riavvolge"
*/
inline bool is_nodo_in_memoria(unsigned long nodo)
{
  assert(nodo < numero_nodi);

  if(nodo >= primo_nodo_in_memoria) //e' nella parte non-riavvolta
    return (nodo - primo_nodo_in_memoria < quanti_nodi_in_memoria); //e' nell'intervallo giusto?
  else //(nodo < primo_nodo_in_memoria)
    return (nodo <= ultimo_nodo_in_memoria) && (ultimo_nodo_in_memoria < primo_nodo_in_memoria); //e' nella parte riavvolta?
}



//calcola la posizione di 'nodo' all'interno del blocco: ritorna percio' un valore tra 0 e 'capienza_blocco_memoria' - 1
//PRECondizione: nodo deve essere nel blocco
inline unsigned long mappa_nodo_to_pos_blocco(unsigned long nodo)
{
  assert( is_nodo_in_memoria(nodo) );
  long risposta= nodo - primo_nodo_in_memoria;

  if(risposta < 0) //il nodo e' nella parte "riavvolta"
    return (unsigned long) (risposta + numero_nodi);
  else // >= 0 
    return (unsigned long) risposta;
}



//PRECondizioni: entrambi i nodi sono nel blocco in memoria
inline bool precede_in_memoria(unsigned long nodo1, unsigned long nodo2)
{
  assert( is_nodo_in_memoria(nodo1) && is_nodo_in_memoria(nodo2));

  return mappa_nodo_to_pos_blocco(nodo1) < mappa_nodo_to_pos_blocco(nodo2);
}



//visita il nodo passato (last_trovato) e ricorsivamente tutti quei nodi in memoria che sono adiacenti a 'last_trovato', e che sono < 'last_trovato'
//ritorna il numero di nodi raggiunti nella visita che partiva da 'nodo'
unsigned long visita_nodo(unsigned long last_trovato, SwappedStack<unsigned long> * pila)
{
  unsigned long tot_nodi_espansi= 0; //conta quanti nodi abbiamo espanso in questa visita (compreso last_trovato)
  pila->push(last_trovato);

  while( !pila->isEmpty() )
  {
    unsigned long nodo_affiorante;

    bool ris= pila->pop(nodo_affiorante);
    accertati_che( ris, MSG_ERRORE_PILA);

    if( !is_completato(nodo_affiorante) ) //se non e' stato ancora espanso lo facciamo adesso
    {
      set_completato(nodo_affiorante);
      nodi_ancora_da_completare--; //abbiamo completato proprio adesso questo 'nodo': ne resta uno di meno da espandere
      tot_nodi_espansi++;
      
      unsigned long pos_nel_blocco= mappa_nodo_to_pos_blocco(nodo_affiorante);
      unsigned long indice_from= vett_puntatori( pos_nel_blocco ); //estremo inferiore incluso. N.B. (!) Per capire qual'e' il puntatore relativo al nodo bisogna calcolare la posizione del nodo all'interno del blocco
      unsigned long indice_to= vett_puntatori( pos_nel_blocco + 1); //estremo sup. escluso
        
      for(unsigned long indice=indice_from; indice < indice_to ; indice++) //scandiamo la lista di adiacenza del nodo_affiorante
      {
        unsigned long nodo_adiac= vettore_adiac[indice];

        if( !is_completato(nodo_adiac) ) //N.B. questo controllo e' superfluo perche' nelle liste di adiacenza non mettiamo nodi gia' completati..
        {
          bool old_val= set_raggiunto(nodo_adiac); //marchiamo il nodo come raggiunto

          if(!old_val) //se prima non era marcato come raggiunto, allora lo abbiamo raggiunto adesso
            nodi_ancora_da_completare++; //un altro nodo da espandere

          if( is_nodo_in_memoria(nodo_adiac) && precede_in_memoria(nodo_adiac, last_trovato)) //se e' gia' stato completato e' inutile riespanderlo; se non e' in memoria non si puo' espandere adesso ; se viene dopo 'last_trovato' lo troveremo poi con la scansione dei bits
            pila->push(nodo_adiac);
        }
      }
    }
  }

  return tot_nodi_espansi;
}



/*
  visita la porzione di grafo contenuta attualmente nel blocco in memoria (i.e. nelle liste di adiacenza)
  alla fine della visita, ogni nodo rappresentato nel blocco o e' non-raggiunto, oppure e'
  raggiunto e completato.
*/
//ritorna il numero di nodi raggiunti nella visita del blocco
unsigned long visita_blocco_in_memoria()
{
  bool found;
  unsigned long last_trovato;
  unsigned long tot_nodi_raggiunti= 0;
  SwappedStack<unsigned long> pila(DIM_BLOCCO_STACK); //dichiarata qui e non in visita_nodo() per non riallocarla ogni volta
  
  found= nodi_raggiunti->findNext1(last_trovato, primo_nodo_in_memoria); //cerchiamo il primo nodo raggiunto, a partire da 'primo_nodo_in_memoria'
  
  while(found && (last_trovato - primo_nodo_in_memoria < quanti_nodi_in_memoria)) //Finche' troviamo dei nodi da visitare che sono in memoria (notare che la seconda condizione e' la particolarizzione di is_nodo_in_memoria() al caso in cui (nodo >= primo_nodo_in_memoria)). Notare che e' sicuramente last_trovato - primo_nodo_in_memoria >=0 perche' partiamo da primo_nodo_in_memoria a considerare i bits, e in senso crescente.
  {
    if( !is_completato(last_trovato) )
      tot_nodi_raggiunti += visita_nodo(last_trovato, &pila);

    found= nodi_raggiunti->findNext1(last_trovato); //cerca un nodo marcato a partire dal nodo successivo a quello appena trovato
  }
  
  
  if(ultimo_nodo_in_memoria < primo_nodo_in_memoria) //gli adiacenti hanno la parte "riavvolta"
  {
    found= nodi_raggiunti->findNext1(last_trovato, 0); //la parte riavvolta comincia dal nodo 0

    while(found && (last_trovato <= ultimo_nodo_in_memoria)) //Finche' troviamo dei nodi da visitare che sono in memoria (notare che la seconda condizione e' la particolarizzione di is_nodo_in_memoria(), sfruttando le condizioni gia' vere per ipotesi).
    {
      if( !is_completato(last_trovato) )
        tot_nodi_raggiunti += visita_nodo(last_trovato, &pila); //Finche' troviamo dei nodi da visitare che sono in memoria.

      found= nodi_raggiunti->findNext1(last_trovato); //cerca un nodo marcato a partire dal nodo successivo a quello appena trovato
    }    
  }
  
  return tot_nodi_raggiunti;
}



//sceglie un nodo tra quelli ancora non completati: ritorna false se sono stati gia' effettuati tutti i tentativi previsti, true altrimenti
bool scegli_prossimo_nodo_iniziale(unsigned long & next_nodo, BufferLetturaMultifile<unsigned long> * buf_ord)
{
  static unsigned long nodi_scanditi= 0;

  if( !(genera_nodi_random && tentativi_random_illimitati) )
    if(nodi_scanditi >= num_lanci_da_effettuare) //allora controlliamo se abbiamo finito i tentativi
      return false;
  
  if(genera_nodi_random)
  {
     next_nodo= numero_nodi*((float) rand() / (RAND_MAX + 1.0));
     assert(next_nodo<numero_nodi);
  }
  else
  {
    int ris;
    
    ris= buf_ord->get(next_nodo);
    accertati_che(ris, "Error 4: Errore, non e' stato letto il numero previsto di elementi dal mfile di input %s", nome_mfile_nodi_test);
  }

  nodi_scanditi++;
  return true;      
}



/*
  salva nel file specificato i nodi contenuti in 'nodi_raggiunti'.
  Il formato e' o come vettore di bit, o come elenco di nodi (dipende da 'salva_output_come_bits' che raccoglie l'opzione specificata da riga di comando per il salvataggio dei vari insiemi)
*/
void salva_nodi_raggiunti(char * nome_mfile)
{
  if(salva_output_come_bits) //salviamo il vettore di bits proprio come una serie di bit, con l'apposita funzione di VettoreBits. Poiche' pero' ci viene passato un nome di multifile dobbiamo simulare la creazione di un multifile (costituito da un'unica parte)
  {
    const int DIM_BUFFER_NOME= 500;
    char buffer_nomefile[DIM_BUFFER_NOME];

    int car_scritti= snprintf(buffer_nomefile, DIM_BUFFER_NOME, nome_mfile, 0); //creiamo il nome del primo blocco dell'ipotetico multifile
    accertati_che( (car_scritti<DIM_BUFFER_NOME), "Error 5: Nome di file troppo lungo per il buffer usato da salva_nodi_raggiunti():\n `%s'", nome_mfile);
    nodi_raggiunti->salvaSuFile(buffer_nomefile);
  }
  else
  {
    bool found;
    unsigned long nodo_raggiunto;

    //apertura del multifile in cui salvare le informazioni sulla visita:
    eraseMultiFile(nome_mfile); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
    multiFileWriter mfile_out(nome_mfile, 512*1024*1024);
    BufferScritturaMultifile<unsigned long> buf_out(&mfile_out, DIM_BUF_SCRITTURA_OUTPUT);

    //ricerca e output dei nodi marcati come visitati:
    found= nodi_raggiunti->findNext1(nodo_raggiunto, 0); //cerchiamo il primo nodo raggiunto, a partire dal nodo 0

    while(found)
    {
      buf_out.put(nodo_raggiunto);
      found= nodi_raggiunti->findNext1(nodo_raggiunto);
    }

    //chiusura dei multifile:
    buf_out.chiudi(); //N.B. i buffer devono essere chiusi prima dei rispettivi multifile
    mfile_out.close();
  }
}



//aggiunge alla fine del buffer (concatenandoli agli altri) i nodi della componente attuale contenuti in nodi_raggiunti
void salva_nodi_raggiunti_buf(BufferScritturaMultifile<unsigned long> * buf_out)
{
  bool found;
  unsigned long nodo_raggiunto;

  //ricerca e output dei nodi marcati come visitati:
  found= nodi_raggiunti->findNext1(nodo_raggiunto, 0); //cerchiamo il primo nodo raggiunto, a partire dal nodo 0

  while(found)
  {
    buf_out->put(nodo_raggiunto);
    found= nodi_raggiunti->findNext1(nodo_raggiunto);
  }
}



//x fare il debug
//******
void print_adiacenti()
{
  for(unsigned long i=0; i<quanti_nodi_in_memoria ; i++)
  {
    unsigned long degree= vett_puntatori(i+1) - vett_puntatori(i);

    unsigned long true_ID_nodo= i + primo_nodo_in_memoria;
    if(true_ID_nodo >= numero_nodi)
      true_ID_nodo -= numero_nodi; //si "riavvolge" in modulo
    
    printf("* Node %lu (%lu adjacents): ", true_ID_nodo, degree);

    if(degree>0)
    {
      printf("%lu", vettore_adiac[vett_puntatori(i) + 0]);

      for(unsigned long j=1; j<degree; j++) //inseriamo solo marche
        printf(", %lu", vettore_adiac[vett_puntatori(i) + j]);
    }
    printf("\n\n");
  }
}



//*****serve per il debug
void print_raggiunti()
{
  nodi_raggiunti->print("Reached nodes");
}



//*****serve per il debug
void print_completati()
{
  nodi_completati->print("Completed nodes");
}



//alloca una stringa, che va quindi deallocata quando non serve piu'
char * crea_nome_multifile_info(const char * nome_file_nodi)
{
  int lung_nome;
  char * stringa_creata;
  lung_nome= strlen(nome_file_nodi) + strlen(EXT_INFOALBERO);
  stringa_creata= new char[lung_nome + 1]; //un carattere in piu' per il NULL finale
  sprintf(stringa_creata, "%s%s", nome_file_nodi, EXT_INFOALBERO); //concateniamo il nome del file di output principale con un'estensione, ottenendo il nome del file con le info. sugli alberi

  return stringa_creata;
}


//alloca una stringa, che va quindi deallocata quando non serve piu'
char * crea_nome_multifile_tubes(const char * nome_file_nodi)
{
  int lung_nome;
  char * stringa_creata;
  lung_nome= strlen(nome_file_nodi) + strlen(EXT_TUBES);
  stringa_creata= new char[lung_nome + 1]; //un carattere in piu' per il NULL finale
  sprintf(stringa_creata, "%s%s", nome_file_nodi, EXT_TUBES); //concateniamo il nome del file di output principale con un'estensione, ottenendo il nome del file con le info. sugli alberi

  return stringa_creata;
}


//imposta il vettore 'nodi_raggiunti' e la variabile 'nodi_ancora_da_completare' nel caso di una visita con un unico nodo sorgente
void inizializza_raggiunto_singolo_nodo(unsigned long nodo_sorgente)
{
  long errore;
  nodi_raggiunti->reset(); //cancella tutti i bit dei nodi raggiunti
  nodi_raggiunti->set(nodo_sorgente, true, errore); //marchiamo da visitare il nodo iniziale
  accertati_che( (errore==0), MSG_ERRORE_BITS);

  nodi_ancora_da_completare= 1; //abbiamo "raggiunto" un nodo, ma non lo abbiamo ancora espanso  
}



//imposta il vettore 'nodi_raggiunti' e la variabile 'nodi_ancora_da_completare' in base ai successori di nodo_sorgente: tutti i successori non ancora raggiunti vengono marcati come raggiunti e viene aggiornato corripondentemente il numero di nodi che restano da completare
//viene ritornato true se e' stato trovato almeno un successore non-completato, false altrimenti
bool inizializza_raggiunti_successori_nodo(unsigned long nodo_sorgente, bool direzione)
{
  long errore;  
  nodi_raggiunti->reset(); //cancella tutti i bit dei nodi raggiunti
  nodi_ancora_da_completare= 0;

  imposta_direzione_visita(direzione); //impostiamo la direzione della visita
  
  salta_al_nodo( nodo_sorgente, true); //ci posizioniamo sul nodo iniziale eseguendo il rewind
  
  for(unsigned long i=0; i<archi_ultimo_nodo_considerato; i++) //consideriamo la lista di adiacenza del nodo sorgente
  {
    unsigned long adiacente;
    long ris= buf_adiac->get( adiacente );
    accertati_che( ris, MSG_ERRORE_LETTURA_MULTIFILE, grafoADIAC);

    if( !is_completato(adiacente) ) //se invece gia' era completato il nodo destinazione allora l'arco corrispondente e' inutile
    {
      nodi_raggiunti->set(adiacente, true, errore); //marchiamo da visitare il nodo
      accertati_che( (errore==0), MSG_ERRORE_BITS);
      nodi_ancora_da_completare++; //abbiamo "raggiunto" un nodo, ma non lo abbiamo ancora espanso    
    }
  }
  
  return ( nodi_ancora_da_completare > 0 );
}



//effettua una visita nella direzione specificata, iniziando il caricamento degli adiacenti a partire dal nodo specificato
//PREC si assume che i vettori con i bit dei nodi completati e dei nodi raggiunti sia gia' impostato correttamente, cosi' come la variabile 'nodi_ancora_da_completare'
unsigned long esegui_visita(unsigned long nodo_iniziale, bool direzione, unsigned long * nodi_scanditi=NULL)
{
  unsigned long tot_nodi_raggiunti= 0, tot_nodi_caricati= 0;

  imposta_direzione_visita(direzione); //impostiamo la direzione della visita

  salta_al_nodo( nodo_iniziale, true); //ci posizioniamo sul nodo iniziale eseguendo il rewind

  while(nodi_ancora_da_completare > 0)
  {
    tot_nodi_caricati += carica_prossimi_archi();
    
#ifdef VERBOSE_DEBUG
    print_adiacenti();
    printf(" round corrente=%lu\n\n", tot_nodi_caricati/numero_nodi);
#endif

    tot_nodi_raggiunti += visita_blocco_in_memoria();
    
#ifdef VERBOSE_DEBUG
    printf("Dopo la visita in memoria:\n");
    print_raggiunti();
    print_completati();
    printf("\nnodi restanti da espandere:%lu\n", nodi_ancora_da_completare);
#endif

    if(nodi_ancora_da_completare > 0) //se dobbiamo ripetere il loop
    {
      unsigned long dest_salto;
      bool found= prossimo_nodo_raggiunto_non_completato(dest_salto);
      accertati_che(found, "Error 6: ci si aspettava di trovare almeno un nodo raggiunto, ma non completato");
      salta_al_nodo( dest_salto);
    }
  }

  if(nodi_scanditi != NULL)
    (*nodi_scanditi)= tot_nodi_caricati;
    
  return tot_nodi_raggiunti;
}



/*
  Calcola la regione OUT assumendo di conoscere la regione SCC
  Si specifica con 'size_SCC_max' il numero di nodi contenuti in SCC e va
  specificato se occorre o meno salvare su file l'insieme OUT calcolato
*/
void ricava_OUT_da_SCC(unsigned long size_SCC_max, bool salva_su_file)
{
  unsigned long primo_nodo_di_SCC, dim_tot_OUT_SCC;
  printf("Visit OUT, starting from SCC...\n"); fflush(stdout);

  nodi_completati->reset();
  nodi_raggiunti->caricaDaFile(nome_filetmp_max_SCC); //devo espandere tutti i nodi in SCC
  nodi_ancora_da_completare= size_SCC_max; //devo espandere tutti i nodi in SCC
  bool found= nodi_raggiunti->findNext1(primo_nodo_di_SCC, 0); //cerchiamo il nodo da espandere per primo
  accertati_che(found, "Error 7: nessun nodo trovato in SCC");
  dim_tot_OUT_SCC= esegui_visita(primo_nodo_di_SCC, true); //visita in avanti

  //a questo punto ho 'nodi_raggiunti' == 'nodi_completati' == SCC + OUT
  nodi_completati->caricaDaFile(nome_filetmp_max_SCC); //carichiamo SCC in 'nodi_completati'
  nodi_raggiunti->XOR( *nodi_completati ); //togliamo i nodi comuni (quelli di SCC) e otteniamo cosi' OUT

#ifdef VERBOSE_DEBUG
  printf(MSG_DONE);
#endif
  size_Set_OUT= dim_tot_OUT_SCC - size_SCC_max;

  if(salva_su_file)
    salva_nodi_raggiunti(nome_mfile_set_OUT); //salviamo il contenuto di 'nodi_raggiunti'

#ifdef VERBOSE_DEBUG
  printf("OUT found (%lu nodes)\n", size_Set_OUT);
#endif
}



/*
  Calcola la regione IN assumendo di conoscere la regione SCC
  Si specifica con 'size_SCC_max' il numero di nodi contenuti in SCC e va
  specificato se occorre o meno salvare su file l'insieme IN calcolato
*/
void ricava_IN_da_SCC(unsigned long size_SCC_max, bool salva_su_file)
{
  unsigned long primo_nodo_di_SCC, dim_tot_IN_SCC;
  printf("Visit IN, starting from SCC...\n"); fflush(stdout);

  nodi_completati->reset();
  nodi_raggiunti->caricaDaFile(nome_filetmp_max_SCC); //devo espandere tutti i nodi in SCC
  nodi_ancora_da_completare= size_SCC_max; //devo espandere tutti i nodi in SCC
  bool found= nodi_raggiunti->findNext1(primo_nodo_di_SCC, 0); //cerchiamo il nodo da espandere per primo
  accertati_che(found, "Error 7: nessun nodo trovato in SCC");
  dim_tot_IN_SCC= esegui_visita(primo_nodo_di_SCC, false); //visita all'indietro

  //a questo punto ho 'nodi_raggiunti' == 'nodi_completati' == SCC + IN
  nodi_completati->caricaDaFile(nome_filetmp_max_SCC); //carichiamo SCC in 'nodi_completati'
  nodi_raggiunti->XOR( *nodi_completati ); //togliamo i nodi comuni (quelli di SCC) e otteniamo cosi' IN
    
#ifdef VERBOSE_DEBUG
  printf(MSG_DONE);
#endif
  size_Set_IN= dim_tot_IN_SCC - size_SCC_max;

  if(salva_su_file)
    salva_nodi_raggiunti(nome_mfile_set_IN); //salviamo il contenuto di 'nodi_raggiunti'

#ifdef VERBOSE_DEBUG
  printf("IN found (%lu nodes)\n", size_Set_IN);
#endif
}



/*
  ricava, supponendo di conoscere gia' IN, gli elementi di TENDRILS raggiungibili partendo da IN
  vengono specificati i file che contengono IN ed OUT sotto forma di vettori di bits (l'insieme OUT serve per escludere i nodi OUT raggiunti attraverso un tube)
*/
void ricava_TENDRILS_di_IN(const char * nome_file_IN, const char * nome_file_OUT)
{
  unsigned long primo_nodo_di_IN, dim_tot_TENDRILSIN_IN= 0;
  printf("Visit TENDRILS, starting from IN...\n"); fflush(stdout);

  if(size_Set_IN > 0)
  {
    nodi_completati->caricaDaFile(nome_filetmp_max_SCC); //evitiamo di finire in SCC
    nodi_raggiunti->caricaDaFile(nome_file_IN); //carico IN

    nodi_ancora_da_completare= size_Set_IN; //devo espandere tutti i nodi in IN
    bool found= nodi_raggiunti->findNext1(primo_nodo_di_IN, 0); //cerchiamo il nodo da espandere per primo
    accertati_che(found, "Error 8: nessun nodo in IN, ma la dim. di IN risulta >0");
    dim_tot_TENDRILSIN_IN= esegui_visita(primo_nodo_di_IN, true); //visita in avanti

    //a questo punto ho 'nodi_raggiunti' == 'nodi_completati' == TENDRILS_IN + IN
    nodi_completati->caricaDaFile(nome_file_IN); //carichiamo IN in 'nodi_completati'
    nodi_raggiunti->XOR( *nodi_completati ); //togliamo i nodi comuni (quelli di IN)

    //adesso dobbiamo evitare di inserire in TENDRILS_IN dei nodi di OUT raggiungibili con un TUBE:
    nodi_completati->caricaDaFile(nome_file_OUT);
    nodi_completati->NOT( *nodi_completati ); //invertiamo OUT
    nodi_raggiunti->AND( *nodi_completati ); //togliamo cosi' i nodi appartenenti ad OUT
  }
  else
    nodi_raggiunti->reset();  //TENDRILSIN vuoto

#ifdef VERBOSE_DEBUG
  printf(MSG_DONE);
#endif
}



/*
  ricava, supponendo di conoscere gia' OUT, gli elementi di TENDRILS che raggiungono OUT
  vengono specificati i file che contengono IN ed OUT sotto forma di vettori di bits (l'insieme IN serve per escludere i nodi IN raggiunti attraverso un tube)
*/
void ricava_TENDRILS_di_OUT(const char * nome_file_IN, const char * nome_file_OUT)
{
  unsigned long primo_nodo_di_OUT, dim_tot_TENDRILSOUT_OUT= 0;
  printf("Visit TENDRILS, starting from OUT...\n"); fflush(stdout);

  if(size_Set_OUT > 0)
  {  
    nodi_completati->caricaDaFile(nome_filetmp_max_SCC); //evitiamo di finire in SCC
    nodi_raggiunti->caricaDaFile(nome_file_OUT); //carico OUT

    nodi_ancora_da_completare= size_Set_OUT; //devo espandere tutti i nodi in IN
    bool found= nodi_raggiunti->findNext1(primo_nodo_di_OUT, 0); //cerchiamo il nodo da espandere per primo
    accertati_che(found, "Error 9: nessun nodo in OUT, ma la dim. di OUT risulta >0");
    dim_tot_TENDRILSOUT_OUT= esegui_visita(primo_nodo_di_OUT, false); //visita all'indietro

    //a questo punto ho 'nodi_raggiunti' == 'nodi_completati' == TENDRILSOUT + OUT
    nodi_completati->caricaDaFile(nome_file_OUT); //carichiamo OUT in 'nodi_completati'
    nodi_raggiunti->XOR( *nodi_completati ); //togliamo i nodi comuni (quelli di OUT)

    //adesso dobbiamo evitare di inserire in TENDRILS_OUT dei nodi di IN raggiungibili con un TUBE:
    nodi_completati->caricaDaFile(nome_file_IN);
    nodi_completati->NOT( *nodi_completati ); //invertiamo IN
    nodi_raggiunti->AND( *nodi_completati ); //togliamo cosi' i nodi appartenenti ad IN
  }
  else
    nodi_raggiunti->reset(); //TENDRILSOUT vuoto

#ifdef VERBOSE_DEBUG
  printf(MSG_DONE);
#endif
}



/*
  ricava l'insieme TENDRILS
  PRECondizioni: sono gia' noti IN, OUT, SCC
  opzioni: occorre salvare l'insieme calcolato su file?
*/
void ricava_TENDRILS(bool salva_su_file)
{
  char * nome_mfile_tubes=NULL;

  if(salva_su_file)
  {
    nome_mfile_tubes= crea_nome_multifile_tubes(nome_mfile_set_TENDRILS);
    //printf(" XXXXXXXXXXXXXXXXXXXXXXXX %s XXXXXXXXXXXXXXXXXXXXXXXXXX \n",nome_mfile_tubes);
  }
  
  ricava_TENDRILS_di_IN(nome_file_tmp_IN, nome_file_tmp_OUT); //nel vettore dei bit raggiunti c'e' il risultato:
  
  nodi_raggiunti->salvaSuFile(nome_file_tmp_TENDRILS_lato_IN); //salva il risultato (che verrebbe sovrascritto)

  ricava_TENDRILS_di_OUT(nome_file_tmp_IN, nome_file_tmp_OUT); //nel vettore dei bit raggiunti c'e' il risultato:
  
  nodi_completati->copia( *nodi_raggiunti ); //spostiamo il secondo risultato (T.OUT) nell'altro vettore  

  if(salva_su_file)
  {
    nodi_raggiunti->caricaDaFile(nome_file_tmp_TENDRILS_lato_IN); //ripristiniamo T.IN
    nodi_raggiunti->AND( *nodi_completati ); //intersezione dei due TENDRILS = TUBEs  
    salva_nodi_raggiunti( nome_mfile_tubes );
  }
  
  nodi_raggiunti->caricaDaFile(nome_file_tmp_TENDRILS_lato_IN); //T.IN  
  nodi_raggiunti->OR( *nodi_completati ); //unione dei due TENDRILS
  if(salva_su_file)
    salva_nodi_raggiunti( nome_mfile_set_TENDRILS ); //salviamo l'insieme dei TENDRILS
  
  if(salva_su_file)
    delete [] nome_mfile_tubes;  

#ifdef VERBOSE_DEBUG
  printf("TENDRILS found.\n");
#endif

}


/*
  Calcola la regione DISCONNECTED
  PRECondizioni: Sono gia' stati calcolati IN,OUT,SCC,TENDRILS
    N.B. assumiamo che 'nodi_raggiunti' contenga gia' i nodi di TENDRILS
*/
void ricava_DISCONNECTED()
{
  printf("Visit DISCONNECTED...\n"); fflush(stdout);
  nodi_completati->caricaDaFile(nome_file_tmp_IN); //carichiamo IN
  nodi_raggiunti->OR( *nodi_completati ); //uniamo IN

  nodi_completati->caricaDaFile(nome_file_tmp_OUT); //carichiamo OUT
  nodi_raggiunti->OR( *nodi_completati ); //uniamo OUT

  nodi_completati->caricaDaFile(nome_filetmp_max_SCC); //carichiamo SCC
  nodi_raggiunti->OR( *nodi_completati ); //uniamo SCC

  //ora 'nodi_raggiunti' contiene l'unione di SCC, IN, OUT, TENDRILS  --> e' cioe' la WCC principale (comp. _debolmente_ connessa)

  nodi_raggiunti->NOT( *nodi_raggiunti );  //invertiamo tutti i bit ottenendo DISCONNECTED
  salva_nodi_raggiunti( nome_mfile_set_DISC );

#ifdef VERBOSE_DEBUG
  printf("DISCONNECTED found.\n");
#endif

}



/*
  Cattura una richiesta di uscita tramite CTRL+C e imposta una variabile corrispondentemente
  La richiesta di uscita sara' soddisfatta al termine del calcolo della SCC corrente
*/
/*
void richiesta_chiusura(int par)
{
  richiesta_controlc= true;
  fprintf(stderr, "\nRicevuto CTRL+C\nLa richiesta di chiusura sara' accolta alla fine dell'iterazione corrente\n");
}
*/



/*
  Controlla se sono verificate le condizioni per uscire dal programma
  Normalmente vengono specificate condizioni ad hoc perche' il calcolo di TUTTE le SCC sarebbe MOLTO laborioso
*/
bool uscita_anticipata(unsigned long numero_SCC_calcolate, unsigned long tot_nodi_ogni_SCC)
{
  bool risposta= false;
  
  //abbiamo calcolato il numero minimo specificato di SCC?
  if( specificato_max_num_SCC && (numero_SCC_calcolate >= numSCC_max) )
  {

#ifdef VERBOSE_DEBUG
    printf("--> Raggiunto il numero minino previsto di SCC (%ld SCC)\n", numSCC_max);
#endif

    risposta= true;
  }

  //le SCC finora calcolate contengono piu' della percentuale specificata di nodi del grafo?
  if(specificato_max_perc && (100.0*(float)tot_nodi_ogni_SCC/numero_nodi >= percNodi_max))
  {

#ifdef VERBOSE_DEBUG
    printf("--> Raggiunta la percentuale minima prevista di nodi (%ld%%)\n", percNodi_max);
#endif

    risposta= true;
  }

  //eccediamo il tempo massimo specificato?
  if(specificato_max_tempo)
  {
    time_t tempo_adesso;
    time(&tempo_adesso);
    if(difftime(tempo_adesso, tempo_inizio)/60 >= tempo_max) //calcoliamo i minuti trascorsi
    {
#ifdef VERBOSE_DEBUG
      printf("--> Raggiunto il tempo massimo specificato (%ld minuti)\n", tempo_max);
#endif
      risposta= true;
    }
  }

  //c'e' stata una richiesta di interruzione con CTRL+C?
  if(richiesta_controlc)
  {

#ifdef VERBOSE_DEBUG
    printf("--> Uscita per richiesta d'interruzione dell'utente\n");
#endif

    risposta= true;
  }

  //calcolate tutte le SCC del grafo?!
  if(tot_nodi_ogni_SCC == numero_nodi)
  {

#ifdef VERBOSE_DEBUG
    printf(" Sono state calcolate tutte le SCC!\n");
#endif

    risposta= true;
  }

  return risposta;
}



//salva i dati sulla regione SCC soltanto (poiche' e' una componente fortemente connessa c'e' anche il file .infotree corrispondente)
void salva_regione_SCC(unsigned long size_SCC_max, BufferScritturaMultifile<unsigned long> * buf_out, BufferScritturaMultifile<struct InfoAlbero> * buf_infotree)
{
  struct InfoAlbero info;

  nodi_raggiunti->caricaDaFile(nome_filetmp_max_SCC); //ricarichiamo da disco la massima SCC trovata
  salva_nodi_raggiunti_buf(buf_out); //salviamo il contenuto di 'nodi_raggiunti'

  info.num_nodi_albero= size_SCC_max;
  info.punt= 0;
  buf_infotree->put(info);
}


    
void print_stato(unsigned long num_nodi_provati, unsigned long size_SCC_max, unsigned long numero_SCC_calcolate, unsigned long perc)
{
  if(genera_nodi_random && tentativi_random_illimitati)
    printf("Max SCC size = %lu; number of SCC found=%lu (%lu%% of nodes)\n\n", size_SCC_max, numero_SCC_calcolate, perc);
  else
    printf("Max SCC size = %lu; number of SCC found=%lu (%lu%% of nodes)\n\n", size_SCC_max, numero_SCC_calcolate, perc);
}



/*
  Calcola per tentativi la massima SCC (ovvero la regione SCC del WebGraph)
  Ogni tentativo richiede la scelta di un nodo casuale, di cui viene calcolata la corrispondente SCC
  mediante una doppia visita "di raggiungibilita'": in avanti e all'indietro.
  Ogni tentativo esamina un nodo che non appartiene alle precedenti SCC calcolate.
  I nodi da provare vengono scelti o in modo casuale, o vengono letti dal file passato dall'utente.
  Il numero di tentativi effettivamente fatti dipende dalle opzioni della riga di comando.
  
  C'e' un'ottimizzazione per il caso di SCC contenenti un unico nodo, per le quali
  non viene effettuata la visita all'indietro, e a volte nemmeno la prima visita in avanti.
  
  Vengono passati i buffer per leggere i nodi da provare e per scrivere tutti i nodi delle SCC calcolate (o della sola SCC massima)
*/
unsigned long the_big_SCC(BufferLetturaMultifile<unsigned long> * buf_nodi_test, BufferScritturaMultifile<unsigned long> * buf_out, BufferScritturaMultifile<struct InfoAlbero> * buf_infotree)
{
  struct InfoAlbero info_scc;
  unsigned long nodo_iniziale, num_nodi_provati, size_SCC_max, numero_SCC_calcolate, step_incontro_SCC_max, tot_nodi_ogni_SCC;

  //calcolo della prima componente:

  nodi_completati->reset(); //inizializza tutti i bit dei nodi completati a zero

  scegli_prossimo_nodo_iniziale( nodo_iniziale, buf_nodi_test);
  printf("Searching CORE...\n");
  printf("  -> forward visit\n");
  bool raggiunto_qualcosa= inizializza_raggiunti_successori_nodo(nodo_iniziale, true);
  if( raggiunto_qualcosa ) //se non e' stato trovato nemmeno un successore del nodo iniziale non carichiamo niente e non facciamo la visita
    esegui_visita(nodo_iniziale, true); //prima visita (in avanti)

  if( !is_raggiunto(nodo_iniziale) ) //--> la SCC ha un solo nodo
  {
    inizializza_raggiunto_singolo_nodo(nodo_iniziale); //imposta la SCC (il vettore 'nodi_raggiunti') in modo che contenga solo il nodo di partenza
    size_SCC_max= 1; //un solo nodo
    printf("  -> backward visit\n");
  }
  else //la SCC ha uno o piu' nodi
  {
    nodi_completati->NOT( *nodi_raggiunti ); //in modo che figurino come completati i nodi che non sono stati raggiunti in avanti, la cui esplorazione sarebbe inutile (non portano mai a nodi che appartengono alla SCC)
    printf("  -> backward visit\n");
    inizializza_raggiunto_singolo_nodo(nodo_iniziale); //imposta il vettore 'nodi_raggiunti' e la variabile 'nodi_ancora_da_completare'
    size_SCC_max= esegui_visita(nodo_iniziale, false); //la seconda visita (all'indietro) puo' esplorare solo nodi gia' raggiunti dalla prima, quindi tutti i nodi raggiunti dalla seconda visita appartengono alla SCC del nodo_iniziale
  }

  nodi_raggiunti->salvaSuFile(nome_filetmp_max_SCC); //salva questa SCC su file (per mem. la massima SCC trovata finora)
  nodi_raggiunti->salvaSuFile(nome_filetmp_tutte_le_SCC); //salva questa SCC su file (per mem. tutti nodi di tutte le SCC trovate finora)
  nodi_completati->copia( *nodi_raggiunti ); //inizializziamo i bit completati con la SCC appena trovata (e' anche l'insieme di tutte le SCC trovate finora)

  numero_SCC_calcolate= 1;
  num_nodi_provati= 1;
  tot_nodi_ogni_SCC= size_SCC_max;
  step_incontro_SCC_max= 1; //e' la SCC incontrata al primo tentativo

  if( salva_tutte_le_SCC)
  {
    salva_nodi_raggiunti_buf(buf_out);  
    info_scc.punt= 0;
    info_scc.num_nodi_albero= size_SCC_max;
    buf_infotree->put(info_scc);
  }
    
  print_stato(num_nodi_provati, size_SCC_max, numero_SCC_calcolate, 100.0*(float)tot_nodi_ogni_SCC/numero_nodi);

  //calcolo di altre SCC:
  
  while( scegli_prossimo_nodo_iniziale( nodo_iniziale, buf_nodi_test) && !uscita_anticipata(numero_SCC_calcolate, tot_nodi_ogni_SCC) ) //finche' restano nodi da provare
  {
    num_nodi_provati++;

    if( !is_completato(nodo_iniziale) ) //se appartiene ad una delle SCC gia' trovate e' inutile calcolare la sua SCC
    {
      unsigned long size_SCC_corrente;

      //nodi_completati sono gia' stati inizializzati alla fine dell'iterazione precedente in modo da contenere i nodi di tutte le SCC trovate finora, che non vanno piu' esplorati
      printf("  ->forward visit\n");
      bool raggiunto_qualcosa= inizializza_raggiunti_successori_nodo(nodo_iniziale, true);
      if( raggiunto_qualcosa ) //se non e' stato trovato nemmeno un successore del nodo iniziale non carichiamo niente e non facciamo la visita
        esegui_visita(nodo_iniziale, true); //prima visita (in avanti)
        
      if( !is_raggiunto(nodo_iniziale) ) //--> la SCC ha un solo nodo
      {
        inizializza_raggiunto_singolo_nodo(nodo_iniziale); //imposta la SCC (il vettore 'nodi_raggiunti') in modo che contenga solo il nodo di partenza
        size_SCC_corrente= 1; //un solo nodo
        printf("  ->backward visit\n");
      }
      else //la SCC ha uno o piu' nodi
      {
        nodi_completati->NOT( *nodi_raggiunti ); //in modo che figurino come completati i nodi che non sono stati raggiunti in avanti e i nodi delle SCC precedenti, la cui esplorazione sarebbe inutile (non portano mai a nodi che appartengono alla SCC). Il fatto che basti fare "A=NOT B" si deduce dalla tabella della funzione caratteristica da implementare e con la mappa di Karnaugh 1x1 (sfruttando una d.c.c.)
        printf("  ->backward visit\n");
        inizializza_raggiunto_singolo_nodo(nodo_iniziale); //imposta il vettore 'nodi_raggiunti' e la variabile 'nodi_ancora_da_completare'
        size_SCC_corrente= esegui_visita(nodo_iniziale, false); //la seconda visita (all'indietro) puo' esplorare solo nodi gia' raggiunti dalla prima, quindi tutti i nodi raggiunti dalla seconda visita appartengono alla SCC del nodo_iniziale
      }
      
      if(size_SCC_corrente > size_SCC_max)
      {
        size_SCC_max= size_SCC_corrente;
        step_incontro_SCC_max= num_nodi_provati; //ci ricorda a che punto della ricerca e' stata incontrata la SCC piu' grande
        nodi_raggiunti->salvaSuFile(nome_filetmp_max_SCC); //salva questa SCC su file (mem. la massima SCC trovata finora)
      }
      
      if( salva_tutte_le_SCC)
      {
        salva_nodi_raggiunti_buf(buf_out);
        info_scc.num_nodi_albero= size_SCC_corrente;
        info_scc.punt= tot_nodi_ogni_SCC;
        buf_infotree->put(info_scc);
      }
      
      tot_nodi_ogni_SCC += size_SCC_corrente;
      accertati_che( tot_nodi_ogni_SCC<=numero_nodi, "errore: tot_nodi_ogni_SCC > numero_nodi");

      nodi_completati->caricaDaFile(nome_filetmp_tutte_le_SCC); //rilegge tutte le SCC calcolate complessivamente
      nodi_completati->OR( *nodi_raggiunti ); //aggiunge all'insieme dei nodi di tutte le SCC gia' trovate quelli della SCC corrente. Serve sia per salvare su file l'insieme di tutti i nodi che non dobbiamo piu' visitare, sia per inizializzare il vettore in vista della prossima iterazione
      nodi_completati->salvaSuFile(nome_filetmp_tutte_le_SCC); //aggiorna l'insieme dei nodi di tutte le SCC calcolate finora

      numero_SCC_calcolate++;

      print_stato(num_nodi_provati, size_SCC_max, numero_SCC_calcolate, 100.0*(float)tot_nodi_ogni_SCC/numero_nodi);
    } //fine blocco if
  }

#ifdef VERBOSE_DEBUG
  if(genera_nodi_random && tentativi_random_illimitati)
    printf("\n> Calcolo effettuato.\n> La piu' grande SCC trovata ha %lu nodi ed e' stata incontrata al tentativo %lu\n> Sono state calcolate in tutto %lu SCC per un totale di %lu nodi (il %.1f%% dei nodi del grafo).\n", size_SCC_max, step_incontro_SCC_max, numero_SCC_calcolate, tot_nodi_ogni_SCC, 100.0*((float) tot_nodi_ogni_SCC)/numero_nodi);
  else
    printf("\n> Calcolo effettuato.\n> La piu' grande SCC trovata ha %lu nodi ed e' stata incontrata al tentativo %lu/%lu\n> Sono state calcolate in tutto %lu SCC per un totale di %lu nodi (il %.1f%% dei nodi del grafo).\n", size_SCC_max, step_incontro_SCC_max, num_lanci_da_effettuare, numero_SCC_calcolate, tot_nodi_ogni_SCC, 100.0*((float) tot_nodi_ogni_SCC)/numero_nodi);
#endif

  return size_SCC_max;
}



//alloca e apre i buffers; calcola la massima SCC e tutte le regioni del WebGraph richieste dalle opzioni della riga di comando
void effettua_misure()
{
  char * nome_mfile_infotree;
  unsigned long size_SCC_max;
  multiFileReader * mf_nodi_test= NULL;
  BufferLetturaMultifile<unsigned long> * buf_nodi_test= NULL; //buffer per leggere dal file di ordinamento, se vi e' il file di ordinamento

  eraseMultiFile(nome_mfile_output); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mfile_out(nome_mfile_output, 512*1024*1024);
  BufferScritturaMultifile<unsigned long> buf_out(&mfile_out, DIM_BUF_SCRITTURA_OUTPUT);

  //creiamo il nome del multifile con le info. sugli alberi
  nome_mfile_infotree= crea_nome_multifile_info(nome_mfile_output);
  eraseMultiFile(nome_mfile_infotree); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mfile_infotree(nome_mfile_infotree, 512*1024*1024);
  BufferScritturaMultifile<struct InfoAlbero> buf_infotree(&mfile_infotree, DIM_BUF_SCRITTURA_INFOTREE);
  
  if( !genera_nodi_random ) //apertura opzionale del multifile per leggere dal file di input
  {
    mf_nodi_test= new multiFileReader(nome_mfile_nodi_test, numFileTentativi);
    buf_nodi_test= new BufferLetturaMultifile<unsigned long>(mf_nodi_test, DIM_BUF_LETTURA_FILE_INPUT_NODI); //buffer per leggere il multifile a blocchi
  }

    
  alloca_memoria();

  if(usato_seed) 
    srandom(seed); //usiamo il seme random specificato

  richiesta_controlc= false;
//  signal(SIGINT, richiesta_chiusura);

  time(&tempo_inizio); //salviamo il tempo corrente
  
  size_SCC_max= the_big_SCC(buf_nodi_test, &buf_out, &buf_infotree);

  
  if( !salva_tutte_le_SCC) //abbiamo gia' salvato la SCC massima insieme con le altre gia' incontrate?
  {
    salva_regione_SCC(size_SCC_max, &buf_out, &buf_infotree); //salviamo solo la SCC piu' grande
  }


  if(calcola_IN || calcola_TENDRILS || calcola_DISC) //se e' necessario calcolare IN lo facciamo
  {
    ricava_IN_da_SCC(size_SCC_max, calcola_IN);

    if(calcola_TENDRILS || calcola_DISC) //nei bit raggiunti vi e' memorizzato IN: occorre salvarlo?
      nodi_raggiunti->salvaSuFile(nome_file_tmp_IN);
  }


  if(calcola_OUT || calcola_TENDRILS || calcola_DISC) //se e' necessario calcolare OUT lo facciamo
  {
    ricava_OUT_da_SCC(size_SCC_max, calcola_OUT);

    if(calcola_TENDRILS || calcola_DISC) //nei bit raggiunti vi e' memorizzato OUT: occorre salvarlo?
      nodi_raggiunti->salvaSuFile(nome_file_tmp_OUT);
  }

    
  if(calcola_TENDRILS || calcola_DISC) //calcolare TENDRILS + eventualmente DISCONNECTED ?
  {    
    ricava_TENDRILS(calcola_TENDRILS);
    
    if(calcola_DISC)
      ricava_DISCONNECTED(); //assumiamo che TENDRILS sia gia' dentro il vettore dei nodi raggiunti
    
    remove(nome_file_tmp_TENDRILS_lato_IN);
    remove(nome_file_tmp_IN);
    remove(nome_file_tmp_OUT);    
  }

  
  buf_out.chiudi(); //N.B. i buffer devono essere chiusi prima dei rispettivi multifile
  buf_infotree.chiudi();
  mfile_out.close();
  mfile_infotree.close();
  delete [] nome_mfile_infotree; //cancelliamo la stringa dalla memoria

  if( !genera_nodi_random )
  {
    mf_nodi_test->close();
    delete buf_nodi_test;
    delete mf_nodi_test;
  }
                
  libera_memoria();

  // rimozione dei file temporanei:
  remove(nome_filetmp_max_SCC);
  remove(nome_filetmp_tutte_le_SCC);
}





void stampaHelp(int arg,char** args){
  printf("Programs that searches for the Bow-Tie regions\n");
  printf("It uses a semi-external algorithm that requires at least 2 bits for each graph node\n");
  printf("\nUsage:\n\n");
  printf("%s [options]... <memory> <graphFile%%> <fileSetSCC%%>\n", args[0]);
  printf("<memory>:       available memory in MB\n");
  printf("<graphFile%%>:  input graph multifile\n");
  printf("<fileSetSCC%%>: output multiFile name; it contains the nodes in the CORE (as adjacency list)\n");
  printf("\noptions:\n"
         "    -force               performs the visit even if the main memory is not enough\n"
         "-r, -random[=<NumIterations>]  searches CORE using <NumIterations> random nodes\n"
         "-f, -file=<nodesFile%%>        searches CORE using all the nodes stored in <nodesFile%%>\n"         
         "-s, -seed=<semeRandom>         specifies the seed of the random generator (it must be used with -random option)\n"
         "-I, -IN=<fileSetIN%%>          computes all the nodes that reach the CORE\n"
         "-O, -OUT=<fileSetOUT%%>        computes all the nodes that are reached from the CORE\n"
         "-T, -TENDRILS=<fileSetTENDRILS%%>  computes TENDRILS and TUBES regions\n"
         "-D, -DISC=<fileSetDISC%%>      computes DISCONNECTED region\n"         
         "-b, -bits                      saves nodes of each region(IN, OUT,TENDRILS, DISCONNECTED) as bit vector and not as a list of nodes (default)\n"
         "-a, -all                       saves all the computed SCCs in <fileSetSCC%%> and not only the bigger one\n"
         "-p, -percentMax=<perc>         minimum dimension of acceptable CORE (percent of nodes)\n"
         "-n, -numSCCMax=<num>           maximum number of SCCs to consider\n"
         "-m, -minutesMax=<minutes>      time limit of computation\n"         
         "\n");
  printf("%% implies that the multifile name must contain %%d\n\n");
}


//uscita dal programma per errore
#define ERRORE_PAR(STRINGA) {fprintf(stderr,"%s\nUsare l'opzione -h per l'help\n",STRINGA);EXIT(1);}


int main(int argc,char* argv[])
{
    stampaLicenza(argc,argv);

//    signal(SIGABRT, chiudi_tutto); //permette di uscire ordinatamente in seguito ad una assert() fallita o ad altro errore
    
    //poniamo i valori di default per i flag
    esegui_comunque= false;
    genera_nodi_random= false;
    salva_tutte_le_SCC= false;    
    calcola_IN= false;
    calcola_OUT= false;
    calcola_TENDRILS= false;
    calcola_DISC= false;
    salva_output_come_bits= false;
    usato_seed= false;
    specificato_max_tempo= false;
    specificato_max_perc= false;
    specificato_max_num_SCC= false;
    
    int modalita_scelte= 0; 
    int temp;
    
    struct option long_options[] =
    {
      {"force", no_argument, NULL, '!'},
      {"random", optional_argument, NULL, 'r'},
      {"file", required_argument, NULL, 'f'},
      {"IN", required_argument, NULL, 'I'},
      {"OUT", required_argument, NULL, 'O'},
      {"TENDRILS", required_argument, NULL, 'T'},
      {"DISC", required_argument, NULL, 'D'},
      {"bits", no_argument, NULL, 'b'},
      {"all", no_argument, NULL, 'a'},      
      {"seed", required_argument, NULL, 's'},
      {"percenteMax", required_argument, NULL, 'p'},
      {"numSCCMax", required_argument, NULL, 'n'},
      {"minutesMax", required_argument, NULL, 'm'},
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index= 0;

    int car;

    do //esaminiamo le opzioni
    {
      car= getopt_long_only(argc, argv, "r::f:I:O:T:D:bas:p:n:m:", long_options, &option_index);

      switch(car)
      {
        case 'r':
          modalita_scelte |= 1;
          genera_nodi_random= true;
          if(optarg != NULL)
          {
            tentativi_random_illimitati= false;
            num_lanci_da_effettuare= atol( optarg );
            if(num_lanci_da_effettuare < 1)
              ERRORE_PAR("Random parameter must be >= 1");
          }
          else //non e' specificato il numero di tentativi
          {
            tentativi_random_illimitati= true;
            num_lanci_da_effettuare= 1; //serve solo per non far fallire il controllo alla fine del main()
          }
          break;

        case 's':
          usato_seed= true;
          seed= atol( optarg );
          break;
          
        case '!':
          esegui_comunque= true;
          break;

        case 'f':
          genera_nodi_random= false;
          nome_mfile_nodi_test= optarg;
          if(!estMultiFileName(nome_mfile_nodi_test))
            ERRORE_PAR("multifile name <fileOrd> must contains '%d' ");

          numFileTentativi= contaNumFilesMultiFile(nome_mfile_nodi_test);
          if(numFileTentativi<=0)
            ERRORE_PAR("parameter error (1)");
          num_lanci_da_effettuare= getTotalSizeMultiFiles(nome_mfile_nodi_test)/sizeof(unsigned long);
          
          modalita_scelte |= 2;
          break;

        case 'I':
          calcola_IN= true;
          nome_mfile_set_IN= optarg;
          if(!estMultiFileName(nome_mfile_set_IN))
            ERRORE_PAR("multifile name <fileSetIN%%> must contains '%d' ");
          break;

        case 'O':
          calcola_OUT= true;
          nome_mfile_set_OUT= optarg;
          if(!estMultiFileName(nome_mfile_set_OUT))
            ERRORE_PAR("multifile name <fileSetOUT>  must contains '%d' ");
          break;

        case 'T':
          calcola_TENDRILS= true;
          nome_mfile_set_TENDRILS= optarg;
          if(!estMultiFileName(nome_mfile_set_TENDRILS))
            ERRORE_PAR("multifile name <fileSetTENDRILS> must contains '%d' ");
          break;

        case 'D':
          calcola_DISC= true;
          nome_mfile_set_DISC= optarg;
          if(!estMultiFileName(nome_mfile_set_DISC))
            ERRORE_PAR("multifile name <fileSetDISC> must contains '%d' ");
          break;
                    
        case 'b':
          salva_output_come_bits= true;
          break;

        case 'a':
          salva_tutte_le_SCC= true;
          break;
          
        case 'p':
          specificato_max_perc= true;
          temp= atol(optarg);
          if( (temp < 1) || (temp >100) )
            ERRORE_PAR("Percent of nodes must be in the range 1..100");          
          percNodi_max= temp;
          break;

        case 'n':
          specificato_max_num_SCC= true;
          temp= atol(optarg);
          if(temp < 1)
            ERRORE_PAR("number of SCCs must be >= 1");            
          numSCC_max= temp;
          break;
          
        case 'm':
          specificato_max_tempo= true;
          temp= atol(optarg);
          if(temp < 1)
            ERRORE_PAR("Time in min. must be >= 1");
          tempo_max= temp;
          break;

                                                                                                                        
        case -1: //fine delle opzioni, ignoriamo il valore
          break;

        case '?': //parametro sconosciuto: lo ignoriamo, tanto il messaggio d'errore lo stampa getopt()
          EXIT(1);
          break;

        default:
          fprintf(stderr, "Unexcepted error in parse of command line\n");
          EXIT(1);
      }

    } while(car != -1);

    if( (modalita_scelte==0) || (modalita_scelte==3) )
      ERRORE_PAR("Error: specified both options -random and -file");

    if(usato_seed && !genera_nodi_random)
      ERRORE_PAR("Error: used -seed without -random");

    if(salva_output_come_bits && !(calcola_OUT || calcola_IN || calcola_TENDRILS || calcola_DISC))
      ERRORE_PAR("Error: -bits can be used only with: -IN, -OUT, -TENDRILS, -DISC");

    if( argc - optind != 3) //contiamo quanti parametri che non sono opzioni restano: sono tutti i parametri dall'indice "optind" incluso in poi
      ERRORE_PAR("Invalid number of parameters\n");

    //optind indica il primo dei parametri non-opzioni
    memoria_max=  atol(argv[optind + 0]) * ((long long)1024)*1024; //salva convertendo da MB a byte
    char* grafo=         argv[optind + 1];
    sprintf(grafoINFO,"%s.info",grafo);
    sprintf(grafoPRED,"%s.pred",grafo);
    sprintf(grafoSUCC,"%s.succ",grafo);
    
    //grafoPRED=         argv[optind + 2];
    //grafoSUCC=         argv[optind + 3];
    nome_mfile_output= argv[optind + 2];

    //controllo dei multifile passati come argomento:
    if (!estMultiFileName(grafo)) ERRORE_PAR("multifile name <graphFile> must contains '%d' ");
    if (!estMultiFileName(nome_mfile_output)) ERRORE_PAR("multifile name <fileOutput> must contains '%d' ");

    numINFO= contaNumFilesMultiFile(grafoINFO);
    if(numINFO<=0)   		ERRORE_PAR("Graph is empty (INFO field)");

    numSUCC= contaNumFilesMultiFile(grafoSUCC);
    if(numSUCC<=0)
      ERRORE_PAR("Graph is empty (SUCC field)");

    numPRED= contaNumFilesMultiFile(grafoPRED);
    if(numPRED<=0)
      ERRORE_PAR("Graph is empty (PRED field)");

    long long nnodi= getTotalSizeMultiFiles(grafoINFO)/sizeof(struct grafoInfo);
    long long narchi= getTotalSizeMultiFiles(grafoPRED)/sizeof(long);
    
    accertati_che( (narchi==getTotalSizeMultiFiles(grafoSUCC)/sizeof(long)), "Error: invalid IPS format");
    accertati_che( (nnodi < MAX_ULONG), "Too many nodes");
    accertati_che( (narchi < MAX_ULONG), "Too many edges");

    numero_nodi= (unsigned long) nnodi;
    numero_archi= (unsigned long) narchi;

    #ifdef VERBOSE_DEBUG
    printf("\n> Il grafo ha %lu nodi e %lu archi.\n", numero_nodi, numero_archi);
    #endif
//     printf("esegui_comunque=%d\n", esegui_comunque);
//     printf("genera_nodi_random=%d\n", genera_nodi_random);
//     printf("tentativi_random_illimitati=%d\n", tentativi_random_illimitati);

    accertati_che( (num_lanci_da_effettuare>0) && (genera_nodi_random || (num_lanci_da_effettuare<=numero_nodi)), "Invalid number of nodes: %lu\n It must be in the range [1 - %lu].", num_lanci_da_effettuare, numero_nodi - 1);

    effettua_misure();

    EXIT(0);
}
