/*
  Classe che serve per "appoggiare" temporaneamente dei dati di un certo tipo T, per
  recuperarli poi in seguito.
  Se c'e' abbastanza spazio nella memoria allocata dal buffer (la cui dimensione e'
  parametrizzata nel costruttore) allora tutti gli elementi verranno mantenuti in memoria,
  altrimenti i dati verranno depositati di volta in volta in un file temporaneo d'ausilio,
  bufferizzando le scritture grazie proprio alla memoria del buffer.
  Il comportamento della classe si differenzia da quello di una FIFO poiche' nel nostro
  caso prima devono avvenire tutte le scritture nel buffer e poi tutte le letture dal buffer.
  Ovvero quando inizio a togliere elementi dal buffer devo svuotarlo interamente prima di
  poter riprendere ad aggiungere elementi. Cio' viene fatto per ragioni di efficienza, in
  particolare per poter rileggere sequenzialmene il file temporaneo (bufferizzando le letture)
  grazie alla porzione di memoria allocata.


  Ernesto Ricci
  26/01/2003
*/


#ifndef BUFFERAPPOGGIO_H_
#define BUFFERAPPOGGIO_H_

#include <stdio.h>


template <class T>
class BufferAppoggio
{
  public:
    BufferAppoggio (); //crea un buffer usando una dimensione standard per la memoria
    BufferAppoggio (long size); //crea un buffer specificando la quantita' di memoria da allocare
    BufferAppoggio(long dim_memoria, T * buffer_passato); //questa versione usa il buffer passato
    ~BufferAppoggio();
    
    bool depositaValore(T); //inserisce un valore nel buffer: restitusce true in caso di inserimento avvenuto, false altrimenti (in quest'ultimo caso l'inserimento non è avvenuto perché si era già iniziato a svuotare il buffer)
    bool recuperaValore(T &); //estrae un valore dal buffer: restitusce false se non ci sono piu' elementi, true altrimenti
    long getSize() { return elementi_totali;} //numero totale di elementi depositati
    bool inserimentiPossibili() { return modo_riempimento;} //ritorna true se posso inserire nuovi elementi, viceversa ritorna false se ho gia' iniziato a svuotare il buffer: in questo caso posso riprendere gli inserimenti solo dopo aver svuotato completamente il buffer
    void svuota(); //scarta tutto il contenuto del buffer e lo rende pronto per essere riempito di nuovo
  private:
    enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BUFFER= 1};

    bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
    bool modo_riempimento; //indica se attualmente stiamo lavorando nella modalita' inserimento di elementi nel buffer oppure in modalita' svuotamento del buffer (prelievo sequenziale del suo contenuto, a partire dal primo elemento inserito)
    FILE * fp; //handler file temporaneo d'ausilio: se NULL vuol dire che non e' stato necessario finora ricorrere al disco
    T * buffer; //memoria allocata per contenere valori inseriti: la sua dimensione e' stabilita con la chiamata del costruttore, se dovesse risultare insufficiente si fara' ricorso ad un file temporaneo d'ausilio
    long num_elementi_in_memoria; //quanti elementi del buffer sono mantenuti in memoria
    long capienza_buffer; //numero massimo di dati contenuti nella parte di buffer in memoria
    long elementi_totali; //elementi totali nel buffer (memoria + disco)
    long prossimo_da_leggere_in_memoria; //variabile usata nella fase di lettura dal buffer: indica il prossimo elemento in memoria a restituire con la prossima lettura

    void init(long dim_memoria); //invocata dai costruttori
    void init(long dim_memoria, T * buffer_passato); //questa versione usa il buffer passato

    BufferAppoggio & operator=(const BufferAppoggio &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    BufferAppoggio(const BufferAppoggio &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)

};




//dimensione standard per la memoria
template<class T>
BufferAppoggio<T>::BufferAppoggio()
{
  // AG_CHANGE

  // stato_apertura_ok= init(DIM_STANDARD);
}


//specifichiamo la dimensione di memoria utilizzata per la parte in memoria del buffer (la dimensione effettiva puo' essere leggermente inferiore)
template<class T>
BufferAppoggio<T>::BufferAppoggio(long dim_memoria)
{
  init(dim_memoria);
}


//questa versione usa il buffer passato
template<class T>
BufferAppoggio<T>::BufferAppoggio(long dim_memoria, T * buffer_passato)
{
  init(dim_memoria, buffer_passato);
}


template<class T>
void BufferAppoggio<T>::init(long dim_memoria)
{
  fp= NULL;
  elementi_totali= 0;
  num_elementi_in_memoria= 0;
  modo_riempimento= true; //il buffer e' vuoto, dobbiamo riempirlo

  buffer_nostro= true;

  capienza_buffer= dim_memoria / sizeof(T);

  if(capienza_buffer < CAPIENZA_MINIMA_BUFFER)
    capienza_buffer= CAPIENZA_MINIMA_BUFFER;

  buffer= new T[capienza_buffer];
  if(buffer == NULL)
  {
    fprintf(stderr, "BufferAppoggio: Errore nell'allocazione della memoria.\n");
    abort();
  }
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
void BufferAppoggio<T>::init(long dim_memoria, T * buffer_passato)
{
  fp= NULL;
  elementi_totali= 0;
  num_elementi_in_memoria= 0;
  modo_riempimento= true; //il buffer e' vuoto, dobbiamo riempirlo

  buffer_nostro= false;

  capienza_buffer= dim_memoria / sizeof(T);

  if(capienza_buffer < 1)
  {
    fprintf(stderr, "BufferAppoggio: La dimensione del buffer passato non e' corretta\n");
    abort();
  }

  buffer= buffer_passato;
}


template<class T>
BufferAppoggio<T>::~BufferAppoggio()
{
  if(buffer_nostro)
    delete [] buffer;

  if(fp != NULL)
  {
    fclose(fp);
    //la documentazione di tmpfile() dice che il file creato viene cancellato automaticamente alla chiusura dello stesso
  }
}


template<class T>
void BufferAppoggio<T>::svuota()
{
  if(fp != NULL)
  {
    fclose(fp);
    //la documentazione di tmpfile() dice che il file creato viene cancellato automaticamente alla chiusura dello stesso
  }

  fp= NULL;
  elementi_totali= 0;
  num_elementi_in_memoria= 0;
  modo_riempimento= true; //il buffer e' vuoto, dobbiamo riempirlo
}


template<class T>
bool BufferAppoggio<T>::depositaValore(T val)
{
  if(!modo_riempimento) //se abbiamo iniziato a svuotare il buffer non possiamo aggiungere altri elementi finche' il buffer non e' tornato vuoto
    return false;

  //vedo se il buffer e' pieno e se si lo scrivo su file
  if(num_elementi_in_memoria==capienza_buffer)
  {
    if(fp==NULL) //se non siamo ancora ricorsi al file di appoggio..
    {
      fp= tmpfile(); //..lo apriamo adesso
      if(fp==NULL)
      {
        perror("BufferAppoggio: Errore nella creazione del file temporaneo"); abort();
      }
    }

    int ris= fwrite(buffer, sizeof(T), capienza_buffer, fp); //passiamo l'intero contenuto della nostra memoria su disco
    if(ris != capienza_buffer)
    {
      perror("BufferAppoggio: Errore scrivendo nel file temporaneo"); abort();
    }

    num_elementi_in_memoria=0; //abbiamo appena liberato la nostra memoria
  }

  buffer[num_elementi_in_memoria]= val; //salvataggio in memoria del dato passato
  elementi_totali++;
  num_elementi_in_memoria++;
  return true;
}


template<class T>
bool BufferAppoggio<T>::recuperaValore(T & val)
{
  if(elementi_totali<=0)
    return false;

  if(modo_riempimento) //abbiamo appena iniziato a svuotare il buffer?
  {
    if(fp!=NULL) //se stiamo iniziando adesso a svuotare il buffer e se avevamo usato anche il file d'appoggio
    {
      int ris= fwrite(buffer, sizeof(T), num_elementi_in_memoria, fp); //per come funziona la classe in memoria abbiamo sicuramente degli elementi: li scriviamo su file per poi rileggere il tutto
      if(ris != num_elementi_in_memoria)
      {
        perror("BufferAppoggio: Errore scrivendo nel file temporaneo"); abort();
      }

      ris= fseek( fp, 0L, SEEK_SET); //mi riposiziono all'inizio del file
      if(ris != 0)
      {
        perror("BufferAppoggio: Errore con la fseek"); abort();
      }

      num_elementi_in_memoria= 0; //abbiamo appena passato gli ultimi elementi del buffer su disco. Questo forza la lettura successiva dal file (all'inizio del file)
    }
    else //fp==NULL, file non usato: ci prepariamo a leggere la memoria del buffer (dove sono tutti gli elementi)
      prossimo_da_leggere_in_memoria= 0;

    modo_riempimento= false; //abbiamo iniziato a svuotare il buffer. Possiamo ricominciare a riempirlo solo quando si sara' svuotato completamente
  }

  if(num_elementi_in_memoria == 0) //N.B.: sappiamo che elementi_totali > 0 percio' ci sono dati nel file
  {
    int da_leggere= (elementi_totali >= capienza_buffer) ? capienza_buffer : elementi_totali; //carichiamo per intero la nostra memoria, se possibile. Altrimenti carichiamo solo i dati restanti
    int ris= fread(buffer, sizeof(T), da_leggere, fp);
    if(ris != da_leggere)
    {
      perror("BufferAppoggio: Errore leggendo dal file temporaneo"); abort();
    }
    num_elementi_in_memoria= da_leggere;
    prossimo_da_leggere_in_memoria= 0; //ci posizioniamo all'inizio del blocco appena letto per restituire i prossimi dati
  }

  val= buffer[ prossimo_da_leggere_in_memoria ]; //occorre usare un'altra variabile apposita per sapere dove passiamo a leggere il buffer (in memoria)

  prossimo_da_leggere_in_memoria++;
  num_elementi_in_memoria--;
  elementi_totali--;

  if(elementi_totali == 0) //abbiamo finito di svuotare il buffer?
  {
    if(fp != NULL)
    {
      fclose(fp); //la documentazione di tmpfile() dice che il file creato viene cancellato automaticamente alla chiusura dello stesso
      fp= NULL;
    }

    modo_riempimento= true; //finito lo svuotamento, possiamo ricominciare a riempire
  }

  return true; //successo: elemento tolto dal buffer e copiato nella variabile passata
}


#endif
