/*
  Questa classe definisce un buffer parametrizzato (usando templates)
  per leggere da un multifile (usando un multiFileReader).
  E' possibile leggere dati predefiniti, come ad es. long,
  o anche dati non predefiniti, come ad es. delle struct proprie,
  istanziando la classe con il particolare tipo di dato che
  si desidera leggere.

  N.B. La definizione della classe avviene nello stesso file .h
  della dichiarazione a causa del comportamento in fase
  di compilazione dei template: se non venisse incluso anche
  il codice del template all'interno del file che utilizzera'
  delle variabili di tipo BufferLetturaMultifile, allora non verrebbe
  generato il codice per gestire il particolare tipo usato.
  In questo caso si otterrebbe un errore nella fase di link.

  26/11/2002
  Ernesto Ricci
*/

#ifndef BUFFERLETTURAMULTIFILE_H_
#define BUFFERLETTURAMULTIFILE_H_

#include "../streamFiles/multiFileReader.h"

template <class T>  class BufferLetturaMultifile
{
  public:
  BufferLetturaMultifile();
  BufferLetturaMultifile(multiFileReader * mf);
  BufferLetturaMultifile(multiFileReader * mf, long dim_b);
  BufferLetturaMultifile(multiFileReader * mf, long dim_b, T * buffer_passato); //questa versione usa il buffer passato
  ~BufferLetturaMultifile();

  int get(T & data); //legge un dato del tipo parametrizzato dal buffer
  long capienzaBuffer() { return capienza_buffer; }; //numero di elementi nel buffer allocato
  bool skip(long num_elem_da_saltare); //salta, leggendo il file, il numero di elementi specificato. Ritorna true se abbiamo saltato tutti gli elementi desiderati prima della fine del file, false altrimenti

  private:
  enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BUFFER= 1}; //definizione costanti per la classe

  bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
  long capienza_buffer; //numero massimo di dati contenuti nel buffer
  T * buffer; //puntatore alla memoria allocata
  long next_data; //indice prossimo dato nel buffer
  long dati_letti; //quanti dati sono stati effettivamente messi nel buffer l'ultima volta
  multiFileReader * mfile; //il multifile da leggere

  void init(multiFileReader * mf, long dim_b); //invocata dai costruttori
  void init(multiFileReader * mf, long dim_b, T * buffer_passato); //questa versione usa il buffer passato

  BufferLetturaMultifile & operator=(const BufferLetturaMultifile &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  BufferLetturaMultifile(const BufferLetturaMultifile &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};


//********************** segue la definizione della classe ************************************************************



template<class T>
BufferLetturaMultifile<T>::BufferLetturaMultifile()
{
  init(NULL, DIM_STANDARD/sizeof(T));
}


template<class T>
BufferLetturaMultifile<T>::BufferLetturaMultifile(multiFileReader * mf)
{
  init(mf, DIM_STANDARD/sizeof(T)); //alloca un numero di elementi che si avvicina quanto possibile alla costante fissata
}


template<class T>
BufferLetturaMultifile<T>::BufferLetturaMultifile(multiFileReader * mf, long dim_b)
{
  init(mf, dim_b/sizeof(T)); //alloca un numero di elementi pari alla dimensione richiesta al buffer (approssimando per difetto se necessario)
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
BufferLetturaMultifile<T>::BufferLetturaMultifile(multiFileReader * mf, long dim_b, T * buffer_passato)
{
  init(mf, dim_b/sizeof(T), buffer_passato);
}


template<class T>
BufferLetturaMultifile<T>::~BufferLetturaMultifile()
{
  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] buffer;
}


template<class T>
void BufferLetturaMultifile<T>::init(multiFileReader * mf, long dim_b) //dim_b e' il numero di elementi del tipo generico T che il buffer deve contenere
{
  mfile= mf;
  capienza_buffer= dim_b;
  buffer_nostro= true;

  if(capienza_buffer < CAPIENZA_MINIMA_BUFFER)
    capienza_buffer= CAPIENZA_MINIMA_BUFFER; //forziamo ad aver un numero minimo di elementi nel buffer

  buffer= new T[capienza_buffer]; //alloca spazio per il numero richiesto di elementi del tipo parametrico T
  if(buffer == NULL)
  {
    fprintf(stderr, "BufferLetturaMultifile: Errore nell'allocazione della memoria.\n");
    abort();
  }

  dati_letti= capienza_buffer; //serve per la prima get(): fingiamo di aver gi letto <capienza_buffer> elementi in modo da capire che non stiamo nell'ultimo blocco del file
  next_data= capienza_buffer;//ci posizioniamo alla fine del buffer in modo da forzare il caricamento alla prima richiesta di un dato
}


//questa versione usa il buffer passato
template<class T>
void BufferLetturaMultifile<T>::init(multiFileReader * mf, long dim_b, T * buffer_passato) //dim_b e' il numero di elementi del tipo generico T che il buffer deve contenere
{
  mfile= mf;
  capienza_buffer= dim_b;
  buffer_nostro= false;

  if(capienza_buffer < 1)
  {
    fprintf(stderr, "BufferLetturaMultifile: La dimensione del buffer passato non e' corretta\n");
    abort();
  }

  buffer= buffer_passato; //useremo il buffer specificato

  dati_letti= capienza_buffer; //serve per la prima get(): fingiamo di aver gi letto <capienza_buffer> elementi in modo da capire che non stiamo nell'ultimo blocco del file
  next_data= capienza_buffer;//ci posizioniamo alla fine del buffer in modo da forzare il caricamento alla prima richiesta di un dato
}


template<class T>
int BufferLetturaMultifile<T>::get(T & data)
{
  if(mfile==NULL)
    return 0; //nessun dato restituito

  if(next_data >= dati_letti) //i dati gi caricati nel buffer sono finiti
  {
    if(dati_letti < capienza_buffer) //l'ultima lettura ha raggiunto la fine del file
      return 0; //nessun dato restituito (finito il buffer e finito il file)
    //altrimenti possiamo effettuare una nuova lettura dal file per ricaricare il buffer

    dati_letti= mfile->read( buffer, sizeof(T), capienza_buffer); //riempiamo il buffer di dati
    if(dati_letti < 1)
      return 0; //se era finito cotemporaneamente il buffer ed il file

    next_data= 0; //ci riportiamo all'inizio del buffer
  }
  //restituiamo un dato
  data= buffer[next_data];
  next_data++;
  return 1;
}


//salta, leggendo il file, il numero di elementi specificato
//ritorna true se abbiamo saltato tutti gli elementi desiderati prima della fine del file, false altrimenti
template<class T>
bool BufferLetturaMultifile<T>::skip(long num_elem_da_saltare)
{
  if(mfile==NULL)
    return 0; //nessun dato restituito

  int dati_rimasti= dati_letti - next_data;
  if( num_elem_da_saltare <= dati_rimasti ) //se il numero di posizioni da saltare e' inferiore o uguale al numero di elementi ancora da leggere presenti nel buffer
  {
    next_data += num_elem_da_saltare; //facciamo come se avessimo letto <num_elem_da_saltare> elementi ed usciamo
    return true;
  }
  //else..

  num_elem_da_saltare -= dati_rimasti; //saltiamo intanto i dati che erano rimasti nel buffer
  next_data= dati_letti; //come se li avessimo letti normalmente: ci assicura uscendo se (dati_letti < capienza_buffer) che valga la condizione di EOF sul buffer

  if(dati_letti < capienza_buffer) //l'ultima lettura ha raggiunto la fine del file
    return false; //finito il buffer e finito il file: non si possono leggere altri elementi
  //else..

  int blocks_to_skip= num_elem_da_saltare/capienza_buffer; //numero di blocchi in mezzo da saltare
  if(blocks_to_skip > 0)
  {
/*    for(int i=0; i<blocks_to_skip; i++)
    {
      dati_letti= mfile->read( buffer, sizeof(T), capienza_buffer);
      if(dati_letti<capienza_buffer)
        return false;
    }*/
    long long ris= mfile->seekInAvanti(sizeof(T), capienza_buffer*blocks_to_skip);
    if(ris != capienza_buffer*blocks_to_skip) //seek non eseguita con successo
    {
      //printf("NOOOOO - file BufferLetturaMultiFile.h - \ncapienza_buffer*blocks_to_skip=%ld - ris=%lld\n",capienza_buffer*blocks_to_skip,ris);
      dati_letti= 0; //serve ad assicurare la condizione dati_letti < capienza_buffer, in modo che le eventuali get() successive trovino EOF
      return false; //non siamo riusciti a "leggere" tutti gli elementi che erano da saltare
    }

    num_elem_da_saltare= num_elem_da_saltare % capienza_buffer; //quanti ne restano da saltare
    if(num_elem_da_saltare==0) //abbiamo appena saltato un numero intero di blocchi, non dobbiamo leggere (i.e. saltare) altro
      return true;
  }

  //a questo punto e' sempre: capienza_buffer > num_elem_da_saltare > 0, dati_letti == capienza_buffer
  dati_letti= mfile->read( buffer, sizeof(T), capienza_buffer); //proviamo a riempire il buffer
  next_data= num_elem_da_saltare; //saltiamo i primi elementi del buffer
  return (dati_letti >= num_elem_da_saltare); //abbiamo saltato tutti gli elementi desiderati prima della fine del file?  N.B. se viene restituito false, perche' next_data=num_elem_da_saltare > dati_letti (con dati_letti<num_elem_da_saltare<capienza_buffer) -----> in questo caso anche le get() successive restituiranno false, come e' giusto

}

#endif
