/*
  Coda dinamica di elementi di un certo tipo parametrico. 
  La coda e' composta di blocchi di elementi allocati dinamicamente (di dimensione
  prefissata e definita al momento della creazione della coda) 
  gestiti a loro volta con una politica di coda.
  Ovvero: ogni volta che l'ultimo blocco della coda si riempie completamente
  un nuovo blocco viene allocato e aggiunto alla fine della coda.
  Ogni volta invece che il primo blocco di elementi della coda si svuota esso viene
  deallocato (tranne il caso particolare in cui e' anche l'unico blocco,
  perche' in qusto caso lo lasciamo allocato).
  
  Allocando la memoria a blocchi di elementi, anziche' per singoli elementi,
  si risparmia spazio in puntatori (un puntatore per ogni blocco anziche' ogni elemento)
  e si puo' gestire ogni singolo blocco come un'array circolare.
  In pratica si hanno quasi le stesse prestazioni di una coda statica realizzata
  tramite un unico array circolare, con la possibilita' di espansione dinamica della coda.


  
  Ci sono due versioni differenti, ma molto simili (da scegliere commentando/decommentando
  la linea "#define CONTEGGIO_CENTRALIZZATO 1"):
   
  * Nella seconda versione ogni blocco memorizza la quantita' di elementi allocati in esso
  e c'e' quindi un piccolo spreco di memoria perche' il numero di elementi nei blocchi
  interni (i.e. tutti i blocchi tranne il primo e l'ultimo) e' noto e pari alla capienza
  di ciascun blocco (poiche' i blocchi interni sono sempre pieni).  
  * Viceversa la prima versione memorizza in CodaDinamica il numero degli elementi nel
  primo e nell'ultimo blocco e risparmia cosi' un contatore per ogni blocco.
  
  La seconda versione e' piu' leggibile, dato che la prima deve effettuare
  controlli aggiuntivi per il caso in cui si ha un unico blocco e quindi occorre che
  le due informazioni coincidano. Notare poi che la prima versione risulta davvero
  vantaggiosa solo nel caso in cui ci siano molti blocchi allocati contemporaneamente.


  Ernesto Ricci
  31/12/2002

*/





#ifndef CODADINAMICA_H_
#define CODADINAMICA_H_

#include <stdio.h>


#define CONTEGGIO_CENTRALIZZATO 1



#ifdef CONTEGGIO_CENTRALIZZATO
/*============================================================================================

  Prima versione: memorizziamo solo il numero di elementi contenuti nel primo blocco e
  nell'ultimo.

==============================================================================================*/


  template <class T>
  class BloccoCoda //memorizza un blocco di elementi della coda. I blocchi sono collegati tra loro da puntatori defininendo una coda a livello piu' "macroscopico"
  {
    public:
    BloccoCoda(long capienza) { //in realta' tutti i blocchi hanno la stessa capienza, ma questa informazione e' mem. in CodaDinamica
                            vettore_memoria= new T[capienza];
                            if(vettore_memoria==NULL)
                            {
                              perror("Errore nell'allocare la memoria per la coda\n");
                              EXIT(1);
                            }
                            punt_next= NULL; //il nuovo blocco allocato andra' per ultimo nella "coda di blocchi"
                          };
    ~BloccoCoda() { delete [] vettore_memoria; }

    T * vettore_memoria; //puntatore alla memoria allocata
    BloccoCoda<T> * punt_next; //puntatore al blocco successivo

    private:
    BloccoCoda & operator=(const BloccoCoda &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    BloccoCoda(const BloccoCoda &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  };



  template <class T>
  class CodaDinamica
  {

    public:
    CodaDinamica(long size_blocchi); //specifichiamo la dimensione in byte per ciascun nuovo blocco di elementi che verra' allocato dinamicamente
    ~CodaDinamica(); //provvede a liberare la memoria allocata

    bool isEmpty(); //coda vuota?
    long size(); //numero elementi nella coda

    void inserisci(const T & data_in); //mette 'data_in' all'ultimo posto della coda ; restituisce false se la coda e' vuota, true altrimenti
    bool estrai(T & data_out); //copia in 'data_out' l'elemento all'inizio della coda e lo toglie

    private:
    enum {CAPIENZA_MINIMA_BLOCCHI_CODA= 10}; //definizione costanti per la classe

    BloccoCoda<T> * p_blocco_iniz; //puntatore al primo blocco allocato
    BloccoCoda<T> * p_blocco_fin; //puntatore all'ultimo blocco allocato

    long indice_primo; //indice da cui estrarre il primo elemento inserito nella coda (rispetto al vettore circolare in cui puo' pensarsi il primo blocco allocato)
    long indice_prossimo_inserito; //indice in cui inserire il prossimo elemento da inserire nella coda (rispetto al vettore circolare in cui puo' pensarsi l'ultimo blocco allocato)
    long numel_primoblocco; //numero elementi nel primo blocco (queste due informazioni non ci sono nella seconda versione)
    long numel_ultimoblocco; //numero elementi nell'ultimo blocco: N.B. se c'e' un unico blocco allocato queste due informazioni devono coincidere
    long capienza_blocco_coda; //numero massimo di elementi che un blocco della coda puo' contenere

    CodaDinamica & operator=(const CodaDinamica &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    CodaDinamica(const CodaDinamica &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  };


  //************** segue la definizione della classe ******************



  template<class T>
  CodaDinamica<T>::CodaDinamica(long size_blocchi)
  {
    //printf("\nconteggio centralizzato\n"); 
    capienza_blocco_coda= size_blocchi / sizeof(T); //numero elementi corrispondenti a quella dimensione del blocco
    if(capienza_blocco_coda < CAPIENZA_MINIMA_BLOCCHI_CODA)
      capienza_blocco_coda= CAPIENZA_MINIMA_BLOCCHI_CODA;

    p_blocco_iniz= new BloccoCoda<T>(capienza_blocco_coda);
    p_blocco_fin= p_blocco_iniz;
    indice_primo= 0;
    indice_prossimo_inserito= 0;
    numel_primoblocco= 0;
    numel_ultimoblocco= 0;
  }



  //provvede a liberare la memoria allocata
  template<class T>
  CodaDinamica<T>::~CodaDinamica()
  {
    BloccoCoda<T> * aux;
    bool fine= false;

    do
    {
      fine= (p_blocco_iniz == p_blocco_fin);

      aux= p_blocco_iniz;
      p_blocco_iniz= p_blocco_iniz->punt_next;
      delete aux;

    } while( !fine  );

  }


  //coda vuota?
  template<class T>
  bool CodaDinamica<T>::isEmpty()
  {
    return ( (p_blocco_iniz == p_blocco_fin) && (numel_primoblocco == 0) ); //un solo blocco vuoto (perche' se anche non ci sono elementi nella coda manteniamo almeno un blocco)
  }



  //numero elementi nella coda
  template<class T>
  long CodaDinamica<T>::size()
  {
    long tot= numel_primoblocco;

    if(p_blocco_iniz == p_blocco_fin) //se c'e' un unico blocco abbiamo finito cosi'
      return tot;

    //altrimenti ho un primo blocco, 0 o piu' blocchi interni tutti pieni, e il blocco finale
    BloccoCoda<T> * aux= p_blocco_iniz->punt_next; //punta al prossimo blocco da esaminare

    while(aux != p_blocco_fin) //aggiunge gli elementi complessivamente presenti nei blocchi interni
    { //c'e' ancora almeno un altro blocco
      tot += capienza_blocco_coda; // i blocchi interni (i.e. tutti esclusi il primo e l'ultimo) sono completamente pieni
      aux= aux->punt_next; //passa al blocco successivo
    }

    tot += numel_ultimoblocco; //aggiungiamo gli elementi del blocco finale
    return tot;
  }



  //mette 'data_in' all'ultimo posto della coda
  template<class T>
  void CodaDinamica<T>::inserisci(const T & data_in)
  {
    if(numel_ultimoblocco == capienza_blocco_coda) //l'ultimo blocco e' pieno
    {
      p_blocco_fin->punt_next= new BloccoCoda<T>(capienza_blocco_coda); //allochiamo un nuovo blocco
      p_blocco_fin= p_blocco_fin->punt_next; //aggiorna il puntatore al blocco finale
      numel_ultimoblocco= 0; //il nuovo blocco appena creato e' ancora vuoto
    }

    p_blocco_fin->vettore_memoria[indice_prossimo_inserito]= data_in; //memorizza il nuovo dato nella prima posizione libera dell'ultimo blocco

    numel_ultimoblocco++;
    if(p_blocco_iniz == p_blocco_fin) //se il primo e l'ultimo sono lo stesso blocco devo aggiornare entrambi i conteggi
      numel_primoblocco++;

    indice_prossimo_inserito++; //incrementa (modulo capienza_blocco_coda) l'indice della prossima posizione libera
    if(indice_prossimo_inserito >= capienza_blocco_coda)
      indice_prossimo_inserito -= capienza_blocco_coda;

  }



  //copia in 'data_out' l'elemento all'inizio della coda e toglie quest'ultimo dalla coda ; restituisce false se la coda e' vuota, true altrimenti
  template<class T>
  bool CodaDinamica<T>::estrai(T & data_out)
  {
    if( this->isEmpty() )
      return false;
    //else..

    data_out= p_blocco_iniz->vettore_memoria[indice_primo]; //copiamo il primo elemento inserito in coda

    indice_primo++; //incrementa (modulo capienza_blocco_coda) l'indice del primo elemento inserito in coda
    if(indice_primo >= capienza_blocco_coda)
      indice_primo -= capienza_blocco_coda;

    numel_primoblocco--; //aggiorna il numero di elementi nel primo blocco
    if(p_blocco_iniz == p_blocco_fin) //se il primo e l'ultimo sono lo stesso blocco devo aggiornare entrambi i conteggi
      numel_ultimoblocco--; //notare che se ho un unico blocco vuoto non lo dealloco
    else //ho almeno 2 blocchi: se il primo e' vuoto lo rimuovo
    {
      if(numel_primoblocco == 0)
      { //rimozione primo blocco e aggiornamento puntatore iniziale
        BloccoCoda<T> * aux= p_blocco_iniz;
        p_blocco_iniz= p_blocco_iniz->punt_next;
        delete aux;

        if(p_blocco_iniz == p_blocco_fin) //se adesso e' rimasto un unico blocco gli elementi del primo blocco sono gli stessi dell'ultimo..
          numel_primoblocco= numel_ultimoblocco;
        else //..altrimenti il primo blocco e' adesso completamente pieno (fino adesso era un blocco interno)
          numel_primoblocco= capienza_blocco_coda;
      }
    }

    return true;
  }

#else //CONTEGGIO_CENTRALIZZATO non definito

/*============================================================================================

  Seconda versione: in ogni blocco viene memorizzata l'informazione del numero di elementi
  contenuti nel blocco

==============================================================================================*/

  template <class T>
  class BloccoCoda //memorizza un blocco di elementi della coda. I blocchi sono collegati tra loro da puntatori defininendo una coda a livello piu' "macroscopico"
  {
    public:
    BloccoCoda(long capienza) { //in realta' tutti i blocchi hanno la stessa capienza, ma questa informazione e' mem. in CodaDinamica
                            vettore_memoria= new T[capienza];
                            if(vettore_memoria==NULL)
                            {
                              perror("Errore nell'allocare la memoria per la coda\n");
                              EXIT(1);
                            }
                            num_el= 0;
                            punt_next= NULL; //il nuovo blocco allocato andra' per ultimo nella "coda di blocchi"
                          };
    ~BloccoCoda() { delete [] vettore_memoria; }

    T * vettore_memoria; //puntatore alla memoria allocata
    BloccoCoda<T> * punt_next; //puntatore al blocco successivo
    int num_el; //numero elementi nel blocco: N.B. questo contatore non c'e' nell'altra versione
    
    private:
    BloccoCoda & operator=(const BloccoCoda &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    BloccoCoda(const BloccoCoda &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  };


  template <class T>
  class CodaDinamica
  {

    public:
    CodaDinamica(long size_blocchi); //specifichiamo la dimensione in byte per ciascun nuovo blocco di elementi che verra' allocato dinamicamente
    ~CodaDinamica(); //provvede a liberare la memoria allocata

    bool isEmpty(); //coda vuota?
    long size(); //numero elementi nella coda

    void inserisci(T & data_in); //mette 'data_in' all'ultimo posto della coda ; restituisce false se la coda e' vuota, true altrimenti
    bool estrai(T & data_out); //copia in 'data_out' l'elemento all'inizio della coda e lo toglie

    private:
    enum {CAPIENZA_MINIMA_BLOCCHI_CODA= 10}; //definizione costanti per la classe

    BloccoCoda<T> * p_blocco_iniz; //puntatore al primo blocco allocato
    BloccoCoda<T> * p_blocco_fin; //puntatore all'ultimo blocco allocato

    long indice_primo; //indice da cui estrarre il primo elemento inserito nella coda (rispetto al vettore circolare in cui puo' pensarsi il primo blocco allocato)
    long indice_prossimo_inserito; //indice in cui inserire il prossimo elemento da inserire nella coda (rispetto al vettore circolare in cui puo' pensarsi l'ultimo blocco allocato)
    long capienza_blocco_coda; //numero massimo di elementi che un blocco della coda puo' contenere

    CodaDinamica & operator=(const CodaDinamica &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    CodaDinamica(const CodaDinamica &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  };
  

  //************** segue la definizione della classe ******************


  template<class T>
  CodaDinamica<T>::CodaDinamica(long size_blocchi)
  {
    //printf("\nconteggio NON centralizzato\n");
    capienza_blocco_coda= size_blocchi / sizeof(T); //numero elementi corrispondenti a quella dimensione del blocco
    if(capienza_blocco_coda < CAPIENZA_MINIMA_BLOCCHI_CODA)
      capienza_blocco_coda= CAPIENZA_MINIMA_BLOCCHI_CODA;

    p_blocco_iniz= new BloccoCoda<T>(capienza_blocco_coda);
    p_blocco_fin= p_blocco_iniz;
    indice_primo= 0;
    indice_prossimo_inserito= 0;
  }



  //provvede a liberare la memoria allocata
  template<class T>
  CodaDinamica<T>::~CodaDinamica()
  {
    BloccoCoda<T> * aux;
    bool fine= false;

    do
    {
      fine= (p_blocco_iniz == p_blocco_fin);

      aux= p_blocco_iniz;
      p_blocco_iniz= p_blocco_iniz->punt_next;
      delete aux;

    } while( !fine  );

  }


  //coda vuota?
  template<class T>
  bool CodaDinamica<T>::isEmpty()
  {
    return ( (p_blocco_iniz == p_blocco_fin) && (p_blocco_iniz->num_el == 0) ); //un solo blocco vuoto (perche' se anche non ci sono elementi nella coda manteniamo almeno un blocco)
  }



  //numero elementi nella coda
  template<class T>
  long CodaDinamica<T>::size()
  {
    BloccoCoda<T> * aux= p_blocco_iniz;
    int tot= aux->num_el;

    while(aux != p_blocco_fin)
    { //c'e' ancora almeno un altro blocco
      aux= aux->punt_next; //passa al blocco successivo
      tot += aux->num_el;
    }

    return tot;
  }



  //mette 'data_in' all'ultimo posto della coda
  template<class T>
  void CodaDinamica<T>::inserisci(T & data_in)
  {
    if(p_blocco_fin->num_el == capienza_blocco_coda) //l'ultimo blocco e' pieno
    {
      p_blocco_fin->punt_next= new BloccoCoda<T>(capienza_blocco_coda); //allochiamo un nuovo blocco
      p_blocco_fin= p_blocco_fin->punt_next; //aggiorna il puntatore al blocco finale
    }

    p_blocco_fin->vettore_memoria[indice_prossimo_inserito]= data_in; //memorizza il nuovo dato nella prima posizione libera dell'ultimo blocco
    p_blocco_fin->num_el++;

    indice_prossimo_inserito++; //incrementa (modulo capienza_blocco_coda) l'indice della prossima posizione libera
    if(indice_prossimo_inserito >= capienza_blocco_coda)
      indice_prossimo_inserito -= capienza_blocco_coda;

  }



  //copia in 'data_out' l'elemento all'inizio della coda e toglie quest'ultimo dalla coda ; restituisce false se la coda e' vuota, true altrimenti
  template<class T>
  bool CodaDinamica<T>::estrai(T & data_out)
  {
    if( this->isEmpty() )
      return false;
    //else..

    data_out= p_blocco_iniz->vettore_memoria[indice_primo]; //copiamo il primo elemento inserito in coda

    indice_primo++; //incrementa (modulo capienza_blocco_coda) l'indice del primo elemento inserito in coda
    if(indice_primo >= capienza_blocco_coda)
      indice_primo -= capienza_blocco_coda;

    p_blocco_iniz->num_el--;

    if((p_blocco_iniz->num_el == 0) && (p_blocco_iniz != p_blocco_fin)) //se il blocco iniziale e' vuoto e ci sono piu' blocchi allocati --> deallochiamo il blocco iniziale
    {
      BloccoCoda<T> * aux= p_blocco_iniz;
      p_blocco_iniz= p_blocco_iniz->punt_next;
      delete aux;
    }

    return true;
  }

#endif

#endif

