/*
  Questa classe e' un template che realizza una pila, dotata delle operazioni
  classiche di <push>, <pop>, <top of stack>, <la pila e' vuota?>.
  Le dimensioni della pila, i.e. il numero di oggetti contenuti in essa, possono
  essere grandi a piacere. Comunque gli oggetti della classe in ogni momento utilizzeranno
  una quantita' di memoria limitata e non superiore a quella stabilita al momento
  della creazione della pila.
  Cio' e' ottenuto swappando al disco, in un file temporaneo e quando sia necessario,
  porzioni della pila. In particolare quando viene effettuata una Push e la porzione
  di memoria e' piena, meta' della memoria viene salvata in blocco sul file di swap,
  liberando cosi' una pari quantita' di memoria. Se invece una Pop o una "top of stack"
  trovano la porzione di memoria vuota e se ci sono ancora elementi nella pila (swappati sul disco),
  allora un blocco di tali elementi viene letto in memoria, e la memoria viene
  riempita per meta'. Mantenendo dopo ogni lettura/scrittura su disco la memoria occupata
  per meta', e non per intero, si evita che sequenze del tipo Pop-Push-Pop-Push..
  provochino accessi consecutivi al disco.
  Al momento di trasferire in memoria blocchi dal disco questi (ormai non piu' significativi)
  potrebbero essere rimossi dal file troncandolo. Per una questione di efficienza pero'
  il file temporaneo non viene mai accorciato man mano che viene svuotato degli elementi
  della pila, percio' la sua dimensione cresce monotonicamente finche' non viene
  eliminato: infatti una serie di Push successive potrebbero riallungarlo nuovamente
  rendendo vani e inefficienti i troncamenti precedenti.
  La memoria allocata viene gestita come un vettore "circolare" ; il file temporaneo
  e' in pratica una pila di blocchi (di dimensione fissata e pari alla meta'
  della memoria allocata), per quanto detto sopra.

  N.B.
  1) La Pop ha la forma:
  bool pop(T & data_out)      in cui <data_out> e' il dato prelevato dalla pila
  e non la piu' ovvia:
  T & pop()
  per poter trattare con piu' robustezza il caso di Pop da una pila vuota.
  Infatti nel nostro caso possiamo restituire false in tale eventualita';
  nell'altro caso la Pop restituirebbe un valore indefinito senza la possibilita'
  di segnalare al chiamante che tale valore e' privo di significato.
  Lo stesso discorso vale per la "readTop()".

  Il fatto che usiamo un file temporaneo creato dal S.O. ci impedisce di swappare più di 2 GB sul disco

  Ernesto Ricci
  15/12/2002

*/

#ifndef SWAPPEDSTACK_H_
#define SWAPPEDSTACK_H_

#include <stdio.h>


template <class T>  class SwappedStack
{
  public:
  SwappedStack(); //dimensione standard per la memoria
  SwappedStack(unsigned long dim_memoria); //specifichiamo la dimensione di memoria utilizzata dalla pila (la dimensione effettiva puo' essere leggermente inferiore e viene ritornata da memoryBlockSize())
  SwappedStack(unsigned long dim_memoria, T * buffer_passato); //questa versione usa il buffer passato
  ~SwappedStack(); //provvede a liberare la memoria allocata ed elimina, chiudendolo, il file temporaneo (solo se esso e' stato creato)

  inline bool isEmpty(); //pila vuota?
  void svuota(); //svuota rapidamente la pila (senza cioè usare un ciclo di pop() )
  inline unsigned long stackSize(); //numero elementi nella pila
  unsigned long memoryBlockSize(); //ritorna la dimensione in byte del blocco di memoria allocato

  bool readTop(T & data); //copia in 'data' l'elemento in cima allo stack (mantenendolo) ; restituisce false se la pila e' vuota, true altrimenti
  bool readTop(T & data, unsigned long posizioni_dalla_cima); //(!) estensione a readTop(T & data): leggiamo l' n-esimo elemento partendo dalla cima dello stack. Attenzione! è garantita la lettura solo di elementi che distano dalla cima della pila meno di metà capienza della pila stessa. Cioè se ad es. la pila può contenere in memoria max 10 elementi, io potrò leggere il nodo in cima (indice 0), il successivo (indice 1), ed gli altri 3 aventi indici 2,3,4. Altrimenti l'elemento richiesto potrebbe non essere in memoria, ed in tal caso verrebbe restituito false.
  void push(const T & data_in); //mette 'data_in' in cima allo stack
  bool pop(T & data_out); //copia in 'data_out' l'elemento in cima allo stack che viene eliminato ; restituisce false se la pila e' vuota, true altrimenti

  unsigned long capienzaBloccoSwap() { return capienza_blocco_swap;}
  unsigned long numeroAccessiScrittura() { return num_accessi_w; } //ritorna il numero totale di scritture al disco, solo per scopi statistici
  unsigned long dimTotFileSwap() { return lung_file*capienza_blocco_swap*sizeof(T); } //ritorna la dimensione attuale del file di swap in bytes: serve solo per effettuare statistiche

  private:

  enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BLOCCOSWAP= 1}; //definizione costanti per la classe

  bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
  unsigned long num_blocchi_su_disco; //numero blocchi su disco contenenti elementi significativi
  unsigned long num_elementi_in_memoria; //numero elementi della pila in memoria
  unsigned long primo_in_memoria; //l'indice (nel vettore di memoria) dell'elemento piu' vecchio della pila tra quelli presenti in memoria
  unsigned long capienza_blocco_swap; //numero di elementi che compongono un blocco del file di swap (pari a meta' della capienza del vettore di memoria)
  T * vettore_memoria; //puntatore alla memoria allocata: contiene gli elementi piu' recenti tra quelli della pila. N.B.: puo' accadere che temporaneamente il vettore di memoria non contenga nessun elemento, sebbene la pila non sia vuota. Gli elementi sono in tal caso tutti nel file di swap.
  FILE * file_swap; //il file temporaneo

  unsigned long num_accessi_w; //numero accessi in scrittura: serve solo per effettuare statistiche
  unsigned long lung_file; //memorizza la dimensione attuale del file di swap in numero di blocchi: serve solo per effettuare statistiche

  void init(unsigned long dim_memoria); //invocata dai costruttori
  void init(unsigned long dim_memoria, T * buffer_passato); //questa versione usa il buffer passato
  void torna_un_blocco_indietro(); //usa la fseek per posizionarsi un blocco indietro

  SwappedStack & operator=(const SwappedStack &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  SwappedStack(const SwappedStack &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};



//********************** segue la definizione della classe ************************************************************



//dimensione standard per la memoria
template<class T>
SwappedStack<T>::SwappedStack()
{
  init(DIM_STANDARD);
}


//specifichiamo la dimensione di memoria utilizzata dalla pila (la dimensione effettiva puo' essere leggermente inferiore e viene ritornata da memoryBlockSize())
template<class T>
SwappedStack<T>::SwappedStack(unsigned long dim_memoria)
{
  init(dim_memoria);
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
SwappedStack<T>::SwappedStack(unsigned long dim_memoria, T * buffer_passato)
{
  init(dim_memoria, buffer_passato);
}


template<class T>
void SwappedStack<T>::init(unsigned long dim_memoria)
{
  num_accessi_w= 0;
  lung_file= 0;

  file_swap= NULL;
  num_blocchi_su_disco= 0;
  primo_in_memoria= 0;
  num_elementi_in_memoria= 0;

  buffer_nostro= true;

  capienza_blocco_swap= (dim_memoria/2) / sizeof(T); //calcoliamo la meta' della dim. del blocco in memoria, pari a un blocco del file di swap

  if(capienza_blocco_swap < CAPIENZA_MINIMA_BLOCCOSWAP)
    capienza_blocco_swap= CAPIENZA_MINIMA_BLOCCOSWAP;

  vettore_memoria= new T[2 * capienza_blocco_swap]; //avendo prima diviso per due e poi moltiplicato per 2 otteniamo il massimo numero PARI di elementi contenuti nella dimensione specificata
  if(vettore_memoria == NULL)
  {
    fprintf(stderr, "SwappedStack: Errore nell'allocazione della memoria.\n");
    abort();
  }
}


//questa versione usa il buffer passato
template<class T>
void SwappedStack<T>::init(unsigned long dim_memoria, T * buffer_passato)
{
  num_accessi_w= 0;
  lung_file= 0;

  file_swap= NULL;
  num_blocchi_su_disco= 0;
  primo_in_memoria= 0;
  num_elementi_in_memoria= 0;

  buffer_nostro= false;

  capienza_blocco_swap= (dim_memoria/2) / sizeof(T); //calcoliamo la meta' della dim. del blocco in memoria, pari a un blocco del file di swap

  if(capienza_blocco_swap < 1)
  {
    fprintf(stderr, "SwappedStack: La dimensione del buffer passato non e' corretta\n");
    abort();
  }

  vettore_memoria= buffer_passato; //useremo il buffer specificato
}



//provvede a liberare la memoria allocata ed elimina, chiudendolo, il file temporaneo (solo se esso e' stato creato)
template<class T>
SwappedStack<T>::~SwappedStack()
{

  if(file_swap != NULL)
  {
    fclose(file_swap); //la documentazione di tmpfile() dice che il file creato viene cancellato automaticamente al momento della sua chiusura
  }

  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] vettore_memoria;
}



//svuota rapidamente la pila (senza cioè usare un ciclo di pop() )
template<class T>
void SwappedStack<T>::svuota()
{
  num_blocchi_su_disco= 0;
  primo_in_memoria= 0;
  num_elementi_in_memoria= 0;
}


//pila vuota?
template<class T>
bool SwappedStack<T>::isEmpty()
{
  return ( (num_elementi_in_memoria == 0) && (num_blocchi_su_disco == 0) ); //N.B. non basta controllare che non ci siano elementi in memoria, potrebbero essere tutti nel file di swap
}


//numero elementi nella pila
template<class T>
unsigned long SwappedStack<T>::stackSize()
{
  return (num_elementi_in_memoria + num_blocchi_su_disco * capienza_blocco_swap );
}


//ritorna la dimensione in byte del blocco di memoria allocato
template<class T>
unsigned long SwappedStack<T>::memoryBlockSize()
{
  return (2 * capienza_blocco_swap* sizeof(T));
}


//mette 'data_in' in cima allo stack
template<class T>
void SwappedStack<T>::push(const T & data_in)
{

  if(num_elementi_in_memoria == 2*capienza_blocco_swap) //se non c'e' piu' posto in memoria..
  {
    //..allora creiamo spazio in memoria swappando meta' del vettore di memoria sul disco

    if(file_swap == NULL) //se non e' ancora stato creato il file di swap lo creiamo adesso
    {
      file_swap= tmpfile(); //apre un file temporaneo
      if(file_swap==NULL)
      {
        fprintf(stderr, "SwappedStack: Errore nell'apertura del file di swap per lo stack\n");
        abort();
      }
    }

    unsigned long ris= fwrite( vettore_memoria + primo_in_memoria, sizeof(T), capienza_blocco_swap, file_swap);
    if(ris < capienza_blocco_swap)
    {
      perror("SwappedStack: Errore nello scrivere sul file di swap\n");
      abort();
    }

    //a questo punto possiamo riutilizzare la parte di memoria corrispondente al blocco appena salvato su disco
    num_elementi_in_memoria= capienza_blocco_swap; //ovvero dimezziamo la memoria occupata
    primo_in_memoria= capienza_blocco_swap - primo_in_memoria; //inverte il valore di 'primo_in_memoria' da 0 a 'capienza_blocco_swap' o viceversa : i.e. se era primo_in_memoria=0 --> primo_in_memoria=capienza_blocco_swap, altrimenti avviene il contrario

    num_blocchi_su_disco++;
    num_accessi_w++;
    if(num_blocchi_su_disco > lung_file)
      lung_file= num_blocchi_su_disco; //la lunghezza del file di swap cresce monotonicamente

  }


  unsigned long indice_nuovo= primo_in_memoria + num_elementi_in_memoria; //indice del vettore in cui mem. l'elemento passato alla push()
  if(indice_nuovo >= 2*capienza_blocco_swap) //i.e. calcoliamo la somma precedente "modulo 2*capienza_blocco_swap"
    indice_nuovo -= 2*capienza_blocco_swap;

  vettore_memoria[indice_nuovo]= data_in; //copiamo il dato in memoria
  num_elementi_in_memoria++; //incrementiamo il num. di elementi in memoria

}


//usa la fseek per posizionarsi un blocco indietro
template<class T>
void SwappedStack<T>::torna_un_blocco_indietro()
{
    int ris= fseek(file_swap, -capienza_blocco_swap * sizeof(T), SEEK_CUR); //indietreggiamo della dimensione corrispondente ad un blocco sul disco
    if(ris != 0)
    {
      perror("SwappedStack: Errore nello spostarsi con fseek nel file di swap\n");
      abort();
    }
}


//copia in 'data' l'elemento in cima allo stack (mantenendolo) ; restituisce false se la pila e' vuota, true altrimenti
template<class T>
bool SwappedStack<T>::readTop(T & data)
{
  if( isEmpty() )
    return false;
  //else..

  if(num_elementi_in_memoria == 0) //in memoria non ci sono elementi, ma sul disco c'e' almeno un blocco significativo: carichiamo in memoria l'ultimo blocco swappato su disco
  {
    //ci spostiamo indietro di un blocco per poter leggere l'ultimo blocco scritto
    torna_un_blocco_indietro();

    primo_in_memoria= 0; //ricominciamo a riempire il vettore dall'inizio
    unsigned long ris= fread(vettore_memoria, sizeof(T), capienza_blocco_swap, file_swap); //ricominciamo a riempire il vettore dall'inizio
    if(ris < capienza_blocco_swap)
    {
      perror("SwappedStack: Errore nel leggere dal file di swap\n");
      abort();
    }

    num_elementi_in_memoria= capienza_blocco_swap; //con un blocco riempiamo per meta' il vettore in memoria
    num_blocchi_su_disco--; //decrementiamo il numero di blocchi significativi ancora nel file: ne abbiamo appena letto uno

    torna_un_blocco_indietro(); //ci rispostiamo indietro in modo da sovrascrivere il blocco appena letto (quindi non piu' significativo) con la prossima scrittura sul file
  }

  //prendiamo l'ultimo elemento in memoria (che e' l'ultimo dello stack) e ne passiamo una copia al chiamante
  unsigned long indice_top= (primo_in_memoria + num_elementi_in_memoria - 1 );
  if(indice_top >= 2*capienza_blocco_swap) //i.e. calcoliamo la somma precedente "modulo 2*capienza_blocco_swap"
    indice_top -= 2*capienza_blocco_swap;

  data= vettore_memoria[indice_top];
  return true;
}


//copia in 'data' l'elemento in cima allo stack (mantenendolo) ; restituisce false se la pila e' vuota, true altrimenti
//se si specifica il parametro opzionale 'posizioni_dalla_cima' si puo' leggere l'elemento # 'posizioni_dalla_cima' partendo dalla cima dello stack. Per l'elemento affiorante si usa 0, 1 per il secondo elemento, ecc.
//(!) Attenzione! è garantita la lettura solo di elementi che distano dalla cima della pila meno di metà capienza della pila stessa. Cioè se ad es. la pila può contenere in memoria max 10 elementi, io potrò leggere il nodo in cima (indice 0), il successivo (indice 1), ed gli altri 3 aventi indici 2,3,4. Altrimenti l'elemento richiesto potrebbe non essere in memoria, ed inoltre non si potrebbe assicurare di portarlo in memoria caricando un unico blocco dal disco
template<class T>
bool SwappedStack<T>::readTop(T & data, unsigned long posizioni_dalla_cima)
{
  if(posizioni_dalla_cima >= capienza_blocco_swap) //abbiamo per ipotesi la limitazione che readTop() non puo' tornare indietro di una quantita' >= della capienza di un blocco
    return false;

  if(posizioni_dalla_cima >= stackSize()) //ci sono meno di (posizioni_dalla_cima+1) elementi nella pila, la richiesta non puo' essere soddisfatta
    return false;

  //else..

  if(posizioni_dalla_cima >= num_elementi_in_memoria) //in memoria non c'e' l'elemento richiesto, che pero' e' per ipotesi su disco nell'ultimo blocco scritto: carichiamo allora in memoria l'ultimo blocco swappato su disco
  {
    //ci spostiamo indietro di un blocco per poter leggere l'ultimo blocco scritto
    torna_un_blocco_indietro();

    unsigned long pos_inserimento= capienza_blocco_swap - primo_in_memoria; //"flippa": se primo_in_memoria vale 0 viene capienza_blocco_swap, se invece vale capienza_blocco_swap, allora viene 0
    unsigned long ris= fread(vettore_memoria + pos_inserimento, sizeof(T), capienza_blocco_swap, file_swap); //ricominciamo a riempire il vettore dall'inizio
    if(ris < capienza_blocco_swap)
    {
      perror("SwappedStack: Errore nel leggere dal file di swap\n");
      abort();
    }

    num_elementi_in_memoria += capienza_blocco_swap; //con il blocco letto riempiamo meta' del vettore in memoria
    num_blocchi_su_disco--; //decrementiamo il numero di blocchi significativi ancora nel file: ne abbiamo appena letto uno

    torna_un_blocco_indietro(); //ci rispostiamo indietro in modo da sovrascrivere il blocco appena letto (quindi non piu' significativo) con la prossima scrittura sul file
  }


  //prendiamo l'elemento richiesto in memoria (che e' il # 'posizioni_dalla_cima' partendo dalla cima dello stack) e ne passiamo una copia al chiamante
  unsigned long indice_elem= (primo_in_memoria + num_elementi_in_memoria - 1 - posizioni_dalla_cima);
  if(indice_elem >= 2*capienza_blocco_swap) //i.e. calcoliamo la somma precedente "modulo 2*capienza_blocco_swap"
    indice_elem -= 2*capienza_blocco_swap;

  data= vettore_memoria[indice_elem];
  return true;
}


//copia in 'data_out' l'elemento in cima allo stack che viene eliminato ; restituisce false se la pila e' vuota, true altrimenti
template<class T>
bool SwappedStack<T>::pop(T & data_out)
{
  bool risp= readTop(data_out); //prende l'elemento in cima allo stack

  if(risp)
  {
    num_elementi_in_memoria--; //elimina dallo stack l'ultimo elemento, che era in memoria
    return true;
  }
  else
    return false;
}


#endif

