/*

  Programma per calcolare un sottografo a partire da un grafo specificato (entrambi nel formato .ips).
  Il sottografo e' specificato tramite l'insieme dei suoi nodi, cioe' di tutti quei nodi del grafo
  di partenza che continuano ad appartenere al sottografo, in due possibili modi:
  o come vettore di bits (secondo la convenziona usata dalla classe VettoreBits);
  o come lista di identificatori di nodi.
  E' possibile inoltre scegliere se i nodi specificati devono appartenere al sottografo,
  o se invece sono quelli da escludere.

  La creazione del sottografo avviene con una sola scansione dei file .info, .pred, .succ,
  creando nel frattempo i corrispondenti per il sottografo.
  Usiamo percio' 3 buffers di lettura e 3 di scrittura.
  (in realta' basterebbero i 3 buffers di lettura, il cui contenuto viene poi modificato
   e dato in output per il sottografo, impiegando cosi' meglio lo spazio...)
  La creazione del sottografo comporta due capacita':
  - conoscere se un nodo e' selezionato o meno;
  - conoscere il suo nuovo ID nel sottografo, ovvero fare una conversione del tipo:
    ID grafo partenza --> ID sottografo
            
  Sono utilizzati per questo due metodi.
  Supponiamo che ho disponibili L locazioni in memoria per un vettore di interi (32bit),
  ed un grafo di partenza con N nodi:
  
  * metodo "full memory": se L >= N alloco un vettore con un ID per ciascun nodo,
    che indica sia i nodi selezionati, sia il loro ID nel sottografo.
    I nodi esclusi, cioe', sono quelli marcati con un valore "-1" nel vettore, mentre
    per quelli selezionati il valore nel vettore e' proprio il loro ID nel sottografo.
  
  * metodo "semi-external": se L < N suddivido i nodi in blocchi grandi K=N/L e
    la i-esima locazione del vettore (parziale) memorizzera' l'ID del nodo i*K

    gli ID dei generici nodi "u" si calcolano cosi':
    - si vede qual'e' il blocco in cui ricade "u"  --> b= [u / K]
    - si legge la b-esima locazione del vettore e si ha l'ID del nodo b*K
      (se 'b' e' selezionato, ma comunque l'informazione rappresenta il numero di nodi
       selezionati tra quelli con ID originale < b)
    - si scandisce il vettore di bit dalla posizione b*K alla posizione u
      e si conta il numero "c" di nodi inclusi. L'ID del nodo u e' allora
      pari al numero memorizzato nella locazione b, piu' c.
      
    In pratica l'idea e' di calcolare al volo l'ID di ciascun nodo nel sottografo generato,
    basandoci sul vettore di bits che memorizza i nodi selezionati.
    L'ID nel sottografo del nodo 'u' del grafo di partenza e' pari al numero di nodi
    selezionati che precedono 'u' (cioe' hanno l'ID nel grafo originale < u).
    Per rendere piu' veloce il conteggio, evitando una scansione completa nel caso peggiore
    ad ogni ID da convertire, si utilizzano le locazioni del vettore "di lookup".
    I rappresentanti dei nodi, scelti uniformemente nell'intervallo dei nodi, hanno memorizzati
    nel vettore di lookup il numero di nodi selezionati nel sottografo che li precedono
    nella numerazione del grafo di partenza.
    In questo modo, nel caso peggiore, per una conversione di ID il vettore di bits va scandito
    solo nell'intervallo di nodi tra due rappresentanti: si vede quanti nodi selezionati precedono
    il rappresentante del nodo da convertire, e si aggiungono a questi i nodi selezionati presenti
    tra il rappresentante ed il nodo da convertire.


  Ernesto Ricci
   7/4/2003
        
*/



#include <getopt.h>
#include "main.h"
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include "comuni.h"
#include "../common/scc_utl/VettoreBits.h"
#include "full_mem.h"
#include "semi_ext.h"



unsigned long numero_nodi; //numero nodi del grafo
unsigned long numero_archi; //numero archi del grafo



//opzioni riga di comando:
char grafoINFO[1000];
char grafoPRED[1000];
char grafoSUCC[1000];
char * nome_mfile_nodi_input;
char * prefisso_nomi_mfile_output;
long long memoria_max= -1; //memoria massima cha il programma puo' usare
bool esegui_comunque; //flag per eseguire incondizionatamente la visita anche se la memoria necessaria ?maggiore di quella disponibile
bool nodi_input_come_bits;
bool escludi_nodi_specificati;

//numero parti dei vari multifile di input:
long numINFO;
long numPRED;
long numSUCC;
long numPartiNodiInput;

bool full_memory; //indica se usiamo tutta la memoria che ci serve (4 byte per nodo), oppure se usiamo l'approccio semi-external
unsigned long memoria_per_buffers;

//puntatori alle strutture in memoria:
unsigned long * vettore_Id_nodi= NULL;

unsigned long numero_nodi_selezionati;






/*
void chiudi_tutto(int par) //intercetta la richiesta di chiudere forzatamante il programma da parte di assert() e libera la memoria uscendo ordinatamente
{
  printf("\n Deallocazione della memoria e preparazione all'uscita forzata\n");
  libera_memoria();
  ABORT();
}
*/



//scrive un messaggio ed esce dopo liberando le strutture di memoria impegnate
//per quanto riguarda i parametri usa la stessa sintassi di printf:  printf(const char *formato, ...)
void gestione_errore(const char *formato, ...)
{
  va_list ap;
  //stampiamo il messaggio d'errore ed usciamo:
  fprintf(stderr,"\n");
  va_start (ap, formato); //inizializziamo la struttura 'ap'
  vfprintf (stderr, formato, ap); //scriviamo il messaggio d'errore con tutti i parametri eventuali
  va_end (ap); //chiudiamo con la variabile 'ap'
  fprintf(stderr,"\n");
  perror("");

  libera_memoria();
  ABORT();
}


void indiceProgressione(const char* testo)
{
  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  printf("\r [%c] %s\r", cont, testo);
  fflush(stdout);
}


void indiceProgressione(const char* testo,long perc1,long perc2){

  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  if (perc1<0) perc1=0;
  else if (perc1>100) perc1=100;

  if (perc2<0) perc2=0;
  else if (perc2>100) perc2=100;

  perc2=(long) (20.0*perc2/100);
  //indicatore
  char st[50];

  strcpy(st,"____________________");
  for (long i=0;(i<20)&&(i<perc2);i++ )
     st[i]='=';

  //limite a 30 la lunghezza del testo
  char s[55];
  long i=0;
  for (i=0; (testo[i]!=0)&&(i<40);i++)
    s[i]=testo[i];
  for (; (i<40);i++)
    s[i]=' ';
  s[40]=0;


	printf("\r [%c] %3ld%% %20s %s\r",cont,perc1,st, s);
  fflush(stdout);
}


void eraseMultiFile(const char* nomeMulti){

    long conta=0;

    FILE* fp;
    char nomeFile[1000];

    while(true){
      sprintf(nomeFile,nomeMulti,conta);
      fp=fopen(nomeFile,"r");
      if (fp!=NULL){
          conta++;
          fclose(fp);
          remove(nomeFile);
      }
      else
          return;
    }
}




void alloca_memoria()
{
  memoria_per_buffers= DIM_MINIMA_MEM_BUFFERS;
  long long memoria_ideale= ((long long) numero_nodi)*sizeof(unsigned long) + memoria_per_buffers; //approssimativamente la memoria richiesta

#ifdef TEST_SEMIEXT
  memoria_max= memoria_ideale - 1; //serve solo per forzare l'uso del metodo semi-ext
#endif

  if(memoria_ideale <= memoria_max) //tanta memoria: possiamo usare 4 byte per nodo
  {
    full_memory= true;
    capienza_vettore_lookup= numero_nodi; //per ricordarci che e' il caso full-mem
    memoria_per_buffers += (memoria_max - memoria_ideale); //aggiungo la memoria in piu' ai buffers

    vettore_Id_nodi= new unsigned long[numero_nodi];
    accertati_che( (vettore_Id_nodi != NULL) ,"Errore nell'allocazione della memoria.");    
  }
  else //(memoria_ideale > memoria_max) --> dobbiamo usare un approccio semi-external
  {
    full_memory= false; //indica approccio semi-external
    capienza_vettore_lookup= (unsigned long) (PERC_MIN_RAPPRESENTANTI * numero_nodi); //ci mettiamo dapprima nel caso della richiesta minima
    if(capienza_vettore_lookup == 0)
      capienza_vettore_lookup++;

    long long memoria_prevista_semiext= ((long long) capienza_vettore_lookup)*sizeof(unsigned long) + memoria_per_buffers + numero_nodi/8; //nel caso semi-external occorre anche un vettore di bit, oltre ad un vettore parziale per la conversione ed ai buffers.

    if(memoria_prevista_semiext <= memoria_max) //questa volta siamo dentro le specifiche di memoria o sforiamo ancora?
    {
      capienza_vettore_lookup += (memoria_max - memoria_prevista_semiext) / sizeof(unsigned long); //tutta la memoria in piu' la usiamo per il vettore parziale di conversione.
    }
    else //non e' garantito nemmeno il minimo per l'approccio semi-external: che facciamo? (richiesta all'utente)
    {
      printf("Warning not enought memory (required at least %lld bytes).\n", memoria_prevista_semiext );
      if( !esegui_comunque )
      {
        char car;
        printf("Continue (Y/N)\n");

        car= getchar();
        if((car != 'y') && (car != 'Y'))
        {
          printf("Aborted by used.\n");
          ABORT();
        }
      }
    }

#ifdef TEST_SEMIEXT //*****
  capienza_vettore_lookup= (unsigned long)(0.005f * numero_nodi);
  if(capienza_vettore_lookup == 0) capienza_vettore_lookup++;
#endif

    vettore_Id_nodi= new unsigned long[capienza_vettore_lookup]; //allochiamo un vettore un po' piu' piccolo
    nodi_da_includere= new VettoreBits(numero_nodi);
    accertati_che( (vettore_Id_nodi != NULL) && (nodi_da_includere != NULL) ,"Errore nell'allocazione della memoria.");
  }

/*  
  printf("\n> Memoria allocata: circa %lld KB\n", memoria_max/1024); //allochiamo sempre tutta la memoria che ci viene messa a disposizione
  if(!full_memory)
    printf("> Risorse di memoria minime --> semi-external con un vettore di %lu elementi\n\n", capienza_vettore_lookup);
*/
}



void libera_memoria()
{
  if(vettore_Id_nodi != NULL)
  {
    delete [] vettore_Id_nodi;
    vettore_Id_nodi= NULL;
  }
  
  if(nodi_da_includere != NULL)
  {
    delete nodi_da_includere;
    nodi_da_includere= NULL;
  }  
}



//alloca una stringa, che va quindi deallocata quando non serve piu'
char * crea_nome_multifile_grafo(const char * estensione)
{
  int lung_nome;
  char * stringa_creata;
  
  lung_nome= strlen(prefisso_nomi_mfile_output) + strlen(estensione);
  stringa_creata= new char[lung_nome + 1]; //un carattere in piu' per il NULL finale
  sprintf(stringa_creata, "%s%s", prefisso_nomi_mfile_output, estensione); //concateniamo il nome del file di output principale con un'estensione, ottenendo il nome del file con le info. sugli alberi
  
  return stringa_creata;
}



void estrai_sottografo()
{
  unsigned long num_archi_sel;
  
  alloca_memoria();

  if(full_memory)
  {
    imposta_vettore_nodi();
      
    num_archi_sel= scansione_grafo(); //generiamo il sottografo
  }
  else //approccio semi-external
  {
    imposta_vettore_bits(); //impostiamo i bits dell'approccio semi-external

    calcola_elementi_vettore_lookup();
    
    num_archi_sel= scansione_grafo_semiext(); //generiamo il sottografo
  }
  
#ifdef VERBOSE_DEBUG
  printf("Subgraph size: %lu vertices and %lu edges\n", numero_nodi_selezionati, num_archi_sel);
#endif
  
  libera_memoria();  
}





void stampaHelp(int arg,char** args){
    cout<<"This utility is not designed for stand alone use. It's used as a part of others utilities.\n";
    cout<<"\n";
  printf("Utility to extract a subgraph from an IPS graph\n");
  printf("Required at least %lu MB for buffers and about %.2f bit for each node.\n", DIM_MINIMA_MEM_BUFFERS/(1024*1024), 1 + 32*PERC_MIN_RAPPRESENTANTI);
  printf("\nUsage:\n\n");                                       
  printf("%s [Option]... <memory> <inputGraph%%> <setOfNodes%%> <outputGraph%%>\n", args[0]);
  printf("<memory>:         available memory in MB\n");
  printf("<inputGraph%%>:   input multifile graph name (in IPS format)\n");
  printf("<setOfNodes%%>:   set of nodes to maintain in the output graph\n");
  printf("<outputGraph%%>:  output multifile graph name (in IPS format)\n");

  printf("\nOptions:                                                  \n"
         "    -force        execute program even if memory is not enough \n"
         "-b, -bits         specify that the setOfNodes file is in bit format (1 bit for each node) instead of long format(list of nodes as long)\n"
         "-e, -escludi      specify that nodes in <setOfNodes> are excluded from output graph\n");
         
  printf("%% implies that the multifile name must contains %%d\n");
}


//uscita dal programma per errore
#define ERRORE_PAR(STRINGA) {fprintf(stderr,"%s\nUsare l'opzione -h per l'help\n",STRINGA);EXIT(1);}


int main(int argc,char* argv[])
{
    stampaLicenza(argc,argv);

//    signal(SIGABRT, chiudi_tutto); //permette di uscire ordinatamente in seguito ad una assert() fallita o ad altro errore
    
    //poniamo i valori di default per i flag
    esegui_comunque= false;
    nodi_input_come_bits= false;
    escludi_nodi_specificati= false;
    
    struct option long_options[] =
    {
      {"force", no_argument, NULL, '!'},
      {"bits", no_argument, NULL, 'b'},
      {"escludi", no_argument, NULL, 'e'},      
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index= 0;

    int car;

    do //esaminiamo le opzioni
    {
      car= getopt_long_only(argc, argv, "fbe", long_options, &option_index);

      switch(car)
      {
          
        case '!':
          esegui_comunque= true;
          break;

        case 'b':
          nodi_input_come_bits= true;
          break;

        case 'e':
          escludi_nodi_specificati= true;
          break;
                                                                                                                                                                                                                                                
        case -1: //fine delle opzioni, ignoriamo il valore
          break;

        case '?': //parametro sconosciuto: lo ignoriamo, tanto il messaggio d'errore lo stampa getopt()
          EXIT(1);
          break;

        default:
          fprintf(stderr, "Parameter error\n");
          EXIT(1);
      }

    } while(car != -1);

    if( argc - optind != 4) //contiamo quanti parametri che non sono opzioni restano: sono tutti i parametri dall'indice "optind" incluso in poi
      ERRORE_PAR("Invalid number of parameters.\n");

    //optind indica il primo dei parametri non-opzioni
    memoria_max=  atol(argv[optind + 0]) * ((long long)1024)*1024; //salva convertendo da MB a byte
    char* grafo=argv[optind + 1];
    sprintf(grafoINFO,"%s.info",grafo);
    sprintf(grafoPRED,"%s.pred",grafo);
    sprintf(grafoSUCC,"%s.succ",grafo);
    /*
    grafoINFO=         argv[optind + 1];
    grafoPRED=         argv[optind + 2];    
    grafoSUCC=         argv[optind + 3];
    */
    nome_mfile_nodi_input= argv[optind + 2];    
    prefisso_nomi_mfile_output= argv[optind + 3];

    //controllo dei multifile passati come argomento:
    if (!estMultiFileName(grafoINFO)) ERRORE_PAR("Il nome del multifile di input <grafoINFO> deve contenere una substring '%d' ");
    if (!estMultiFileName(grafoPRED)) ERRORE_PAR("Il nome del multifile di input <grafoPRED> deve contenere una substring '%d' ");
    if (!estMultiFileName(grafoSUCC)) ERRORE_PAR("Il nome del multifile di input <grafoSUCC> deve contenere una substring '%d' ");
    if (!estMultiFileName(nome_mfile_nodi_input)) ERRORE_PAR("Il nome del multifile di output <elencoNodi> deve contenere una substring '%d' ");    
    if (!estMultiFileName(prefisso_nomi_mfile_output)) ERRORE_PAR("Il prefisso <grafoOutput> per i nomi dei multifile nel grafo di output deve contenere una substring '%d' ");

    numINFO= contaNumFilesMultiFile(grafoINFO);
    if(numINFO<=0)   		ERRORE_PAR("Errore nei parametri numero files INFO");

    numSUCC= contaNumFilesMultiFile(grafoSUCC);
    if(numSUCC<=0)
      ERRORE_PAR("Errore nei parametri numero files SUCC");

    numPRED= contaNumFilesMultiFile(grafoPRED);
    if(numPRED<=0)
      ERRORE_PAR("Errore nei parametri numero files PREC");

    numPartiNodiInput= contaNumFilesMultiFile(nome_mfile_nodi_input);
    if(numPartiNodiInput<=0)
      ERRORE_PAR("Errore nei parametri: numero files del multifile <elencoNodi>");
      
    long long nnodi= getTotalSizeMultiFiles(grafoINFO)/sizeof(struct grafoInfo);
    long long narchi= getTotalSizeMultiFiles(grafoPRED)/sizeof(long);
    
    accertati_che( (narchi==getTotalSizeMultiFiles(grafoSUCC)/sizeof(long)), "Errore: le dimensioni dei multifile PRED e SUCC devono coincidere");
    accertati_che( (nnodi < MAX_ULONG), "Troppi nodi per poter usare un unsigned long");
    accertati_che( (narchi < MAX_ULONG), "Troppi archi per poter usare un unsigned long");

    numero_nodi= (unsigned long) nnodi;
    numero_archi= (unsigned long) narchi;

#ifdef VERBOSE_DEBUG
    printf("Graph size: %lu vertices and %lu edges.\n", numero_nodi, numero_archi);
#endif

    estrai_sottografo();

#ifdef TEST_SEMIEXT
printf("\n * Flag TEST_SEMIEXT\n");
printf("\n");
#endif

    EXIT(0);
}

